"""
"""

import copy
from functools import wraps

import numpy as np

from . import markup
from .dimensionality import Dimensionality, p_dict
from .registry import unit_registry
from .decorators import with_doc

PREFERRED = []  # List of preferred quantities for each symbol,
                # e.g. PREFERRED = [pq.mV, pq.pA, pq.UnitQuantity('femtocoulomb', 1e-15*pq.C, 'fC')]
                # Intended to be overwritten in down-stream packages

def validate_unit_quantity(value):
    try:
        assert isinstance(value, Quantity)
        assert value.shape in ((), (1, ))
        assert value.magnitude == 1
    except AssertionError:
        raise ValueError(
                'units must be a scalar Quantity with unit magnitude, got %s'\
                %value
            )
    return value

def validate_dimensionality(value):
    if isinstance(value, str):
        try:
            return unit_registry[value].dimensionality
        except (KeyError, UnicodeDecodeError):
            return unit_registry[str(value)].dimensionality
    elif isinstance(value, Quantity):
        validate_unit_quantity(value)
        return value.dimensionality
    elif isinstance(value, Dimensionality):
        return value.copy()
    else:
        raise TypeError(
            'units must be a quantity, string, or dimensionality, got %s'\
            %type(value)
        )

def get_conversion_factor(from_u, to_u):
    validate_unit_quantity(from_u)
    validate_unit_quantity(to_u)
    from_u = from_u._reference
    to_u = to_u._reference
    assert from_u.dimensionality == to_u.dimensionality
    return from_u.magnitude / to_u.magnitude

def scale_other_units(f):
    @wraps(f)
    def g(self, other, *args):
        other = np.asanyarray(other)
        if not isinstance(other, Quantity):
            other = other.view(type=Quantity)
        if other._dimensionality != self._dimensionality:
            other = other.rescale(self.units, dtype=np.result_type(self.dtype, other.dtype))
        return f(self, other, *args)
    return g

def protected_multiplication(f):
    @wraps(f)
    def g(self, other, *args):
        if getattr(other, 'dimensionality', None):
            try:
                assert not isinstance(self.base, Quantity)
            except AssertionError:
                raise ValueError('can not modify units of a view of a Quantity')
        return f(self, other, *args)
    return g

def check_uniform(f):
    @wraps(f)
    def g(self, other, *args):
        if getattr(other, 'dimensionality', None):
            raise ValueError("exponent must be dimensionless")
        other = np.asarray(other)
        try:
            assert other.min() == other.max()
        except AssertionError:
            raise ValueError('Quantities must be raised to a uniform power')
        return f(self, other, *args)
    return g

def protected_power(f):
    @wraps(f)
    def g(self, other, *args):
        if other != 1:
            try:
                assert not isinstance(self.base, Quantity)
            except AssertionError:
                raise ValueError('can not modify units of a view of a Quantity')
        return f(self, other, *args)
    return g

def wrap_comparison(f):
    @wraps(f)
    def g(self, other):
        if isinstance(other, Quantity):
            if other._dimensionality != self._dimensionality:
                other = other.rescale(self._dimensionality)
            other = other.magnitude
        return f(self, other)
    return g


class Quantity(np.ndarray):

    # TODO: what is an appropriate value?
    __array_priority__ = 21

    def __new__(cls, data, units='', dtype=None, copy=True):
        if isinstance(data, Quantity):
            if units:
                data = data.rescale(units)
            if isinstance(data, unit_registry['UnitQuantity']):
                return 1*data
            return np.array(data, dtype=dtype, copy=copy, subok=True).view(cls)

        ret = np.array(data, dtype=dtype, copy=copy).view(cls)
        ret._dimensionality.update(validate_dimensionality(units))
        return ret

    @property
    def dimensionality(self):
        return self._dimensionality.copy()

    @property
    def _reference(self):
        """The reference quantity used to perform conversions"""
        rq = 1*unit_registry['dimensionless']
        for u, d in self.dimensionality.items():
            rq = rq * u._reference**d
        return rq * self.magnitude

    @property
    def magnitude(self):
        return self.view(type=np.ndarray)

    @property
    def real(self):
        return Quantity(self.magnitude.real, self.dimensionality)

    @real.setter
    def real(self, r):
        self.magnitude.real = Quantity(r, self.dimensionality).magnitude

    @property
    def imag(self):
        return Quantity(self.magnitude.imag, self.dimensionality)

    @imag.setter
    def imag(self, i):
        self.magnitude.imag = Quantity(i, self.dimensionality).magnitude

    @property
    def simplified(self):
        rq = 1*unit_registry['dimensionless']
        for u, d in self.dimensionality.items():
            rq = rq * u.simplified**d
        return rq * self.magnitude

    @property
    def units(self):
        return Quantity(1.0, (self.dimensionality))
    @units.setter
    def units(self, units):
        try:
            assert not isinstance(self.base, Quantity)
        except AssertionError:
            raise ValueError('can not modify units of a view of a Quantity')
        try:
            assert self.flags.writeable
        except AssertionError:
            raise ValueError('array is not writeable')
        to_dims = validate_dimensionality(units)
        if self._dimensionality == to_dims:
            return
        to_u = Quantity(1.0, to_dims)
        from_u = Quantity(1.0, self._dimensionality)
        try:
            cf = get_conversion_factor(from_u, to_u)
        except AssertionError:
            raise ValueError(
                'Unable to convert between units of "%s" and "%s"'
                %(from_u._dimensionality, to_u._dimensionality)
            )
        mag = self.magnitude
        mag *= cf
        self._dimensionality = to_u.dimensionality

    def rescale(self, units=None, dtype=None):
        """
        Return a copy of the quantity converted to the specified units.
        If `units` is `None`, an attempt will be made to rescale the quantity
        to preferred units (see `rescale_preferred`).
        """
        if units is None:
            try:
                return self.rescale_preferred()
            except Exception as e:
                raise Exception('No argument passed to `.rescale` and %s' % e)
        to_dims = validate_dimensionality(units)
        if dtype is None:
            dtype = self.dtype
        if self.dimensionality == to_dims:
            return self.astype(dtype)
        to_u = Quantity(1.0, to_dims)
        from_u = Quantity(1.0, self.dimensionality)
        try:
            cf = get_conversion_factor(from_u, to_u)
        except AssertionError:
            raise ValueError(
                'Unable to convert between units of "%s" and "%s"'
                %(from_u._dimensionality, to_u._dimensionality)
            )
        return Quantity(cf*self.magnitude, to_u, dtype=dtype)

    def rescale_preferred(self):
        """
        Return a copy of the quantity converted to the preferred units and scale.
        These will be identified from among the compatible units specified in the
        list PREFERRED in this module. For example, a voltage quantity might be
        converted to `mV`:
        ```
        import quantities as pq
        pq.quantity.PREFERRED = [pq.mV, pq.pA]
        old = 3.1415 * pq.V
        new = old.rescale_preferred() # `new` will be 3141.5 mV.
        ```
        """
        units_str = str(self.simplified.dimensionality)
        for preferred in PREFERRED:
            if units_str == str(preferred.simplified.dimensionality):
                return self.rescale(preferred)
        raise Exception("Preferred units for '%s' (or equivalent) not specified in "
                        "quantites.quantity.PREFERRED." % self.dimensionality)

    @with_doc(np.ndarray.astype)
    def astype(self, dtype=None, **kwargs):
        '''Scalars are returned as scalar Quantity arrays.'''
        ret = super(Quantity, self.view(Quantity)).astype(dtype, **kwargs)
        # scalar quantities get converted to plain numbers, so we fix it
        # might be related to numpy ticket # 826
        if not isinstance(ret, type(self)):
            if self.__array_priority__ >= Quantity.__array_priority__:
                ret = type(self)(ret, self._dimensionality, dtype=self.dtype)
            else:
                ret = Quantity(ret, self._dimensionality, dtype=self.dtype)

        return ret

    def __array_finalize__(self, obj):
        self._dimensionality = getattr(obj, 'dimensionality', Dimensionality())

    def __array_prepare__(self, obj, context=None):
        if self.__array_priority__ >= Quantity.__array_priority__:
            res = obj if isinstance(obj, type(self)) else obj.view(type(self))
        else:
            # don't want a UnitQuantity
            res = obj.view(Quantity)
        if context is None:
            return res

        uf, objs, huh = context
        if uf.__name__.startswith('is'):
            return obj
        #print self, obj, res, uf, objs
        try:
            res._dimensionality = p_dict[uf](*objs)
        except KeyError:
            raise ValueError(
                """ufunc %r not supported by quantities
                please file a bug report at https://github.com/python-quantities
                """ % uf
                )
        return res

    def __array_wrap__(self, obj, context=None):
        if not isinstance(obj, Quantity):
            # backwards compatibility with numpy-1.3
            obj = self.__array_prepare__(obj, context)
        return obj

    @with_doc(np.ndarray.__add__)
    @scale_other_units
    def __add__(self, other):
        return super().__add__(other)

    @with_doc(np.ndarray.__radd__)
    @scale_other_units
    def __radd__(self, other):
        return np.add(other, self)
        return super().__radd__(other)

    @with_doc(np.ndarray.__iadd__)
    @scale_other_units
    def __iadd__(self, other):
        return super().__iadd__(other)

    @with_doc(np.ndarray.__sub__)
    @scale_other_units
    def __sub__(self, other):
        return super().__sub__(other)

    @with_doc(np.ndarray.__rsub__)
    @scale_other_units
    def __rsub__(self, other):
        return np.subtract(other, self)
        return super().__rsub__(other)

    @with_doc(np.ndarray.__isub__)
    @scale_other_units
    def __isub__(self, other):
        return super().__isub__(other)

    @with_doc(np.ndarray.__mod__)
    @scale_other_units
    def __mod__(self, other):
        return super().__mod__(other)

    @with_doc(np.ndarray.__imod__)
    @scale_other_units
    def __imod__(self, other):
        return super().__imod__(other)

    @with_doc(np.ndarray.__imul__)
    @protected_multiplication
    def __imul__(self, other):
        return super().__imul__(other)

    @with_doc(np.ndarray.__rmul__)
    def __rmul__(self, other):
        return np.multiply(other, self)
        return super().__rmul__(other)

    @with_doc(np.ndarray.__itruediv__)
    @protected_multiplication
    def __itruediv__(self, other):
        return super().__itruediv__(other)

    @with_doc(np.ndarray.__rtruediv__)
    def __rtruediv__(self, other):
        return np.true_divide(other, self)
        return super().__rtruediv__(other)

    @with_doc(np.ndarray.__pow__)
    @check_uniform
    def __pow__(self, other):
        return np.power(self, other)

    @with_doc(np.ndarray.__ipow__)
    @check_uniform
    @protected_power
    def __ipow__(self, other):
        return super().__ipow__(other)

    def __round__(self, decimals=0):
        return np.around(self, decimals)

    @with_doc(np.ndarray.__repr__)
    def __repr__(self):
        return '%s * %s'%(
            repr(self.magnitude), self.dimensionality.string
        )

    @with_doc(np.ndarray.__str__)
    def __str__(self):
        if markup.config.use_unicode:
            dims = self.dimensionality.unicode
        else:
            dims = self.dimensionality.string
        return '%s %s'%(str(self.magnitude), dims)

    if tuple(map(int, np.__version__.split('.')[:2])) >= (1, 14):
        # in numpy 1.14 the formatting of scalar values was changed
        # see https://github.com/numpy/numpy/pull/9883

        def __format__(self, format_spec):
            ret = super().__format__(format_spec)
            if self.ndim:
                return ret
            return ret + f' {self.dimensionality}'

    @with_doc(np.ndarray.__getitem__)
    def __getitem__(self, key):
        ret = super().__getitem__(key)
        if isinstance(ret, Quantity):
            return ret
        else:
            return Quantity(ret, self._dimensionality)

    @with_doc(np.ndarray.__setitem__)
    def __setitem__(self, key, value):
        if not isinstance(value, Quantity):
            value = Quantity(value)
        if self._dimensionality != value._dimensionality:
            # Setting `dtype` to 'd' is done to ensure backwards
            # compatibility, arguably it's questionable design.
            value = value.rescale(self._dimensionality, dtype='d')
        self.magnitude[key] = value

    @with_doc(np.ndarray.__lt__)
    @wrap_comparison
    def __lt__(self, other):
        return self.magnitude < other

    @with_doc(np.ndarray.__le__)
    @wrap_comparison
    def __le__(self, other):
        return self.magnitude <= other

    @with_doc(np.ndarray.__eq__)
    def __eq__(self, other):
        if isinstance(other, Quantity):
            try:
                other = other.rescale(self._dimensionality).magnitude
            except ValueError:
                return np.zeros(self.shape, '?')
        return self.magnitude == other

    @with_doc(np.ndarray.__ne__)
    def __ne__(self, other):
        if isinstance(other, Quantity):
            try:
                other = other.rescale(self._dimensionality).magnitude
            except ValueError:
                return np.ones(self.shape, '?')
        return self.magnitude != other

    @with_doc(np.ndarray.__ge__)
    @wrap_comparison
    def __ge__(self, other):
        return self.magnitude >= other

    @with_doc(np.ndarray.__gt__)
    @wrap_comparison
    def __gt__(self, other):
        return self.magnitude > other

    #I don't think this implementation is particularly efficient,
    #perhaps there is something better
    @with_doc(np.ndarray.tolist)
    def tolist(self):
        #first get a dummy array from the ndarray method
        work_list = self.magnitude.tolist()
        #now go through and replace all numbers with the appropriate Quantity
        if isinstance(work_list, list):
            self._tolist(work_list)
        else:
            work_list = Quantity(work_list, self.dimensionality)
        return work_list

    def _tolist(self, work_list):
        for i in range(len(work_list)):
            #if it's a list then iterate through that list
            if isinstance(work_list[i], list):
                self._tolist(work_list[i])
            else:
                #if it's a number then replace it
                # with the appropriate quantity
                work_list[i] = Quantity(work_list[i], self.dimensionality)

    #need to implement other Array conversion methods:
    # item, itemset, tofile, dump, byteswap

    @with_doc(np.ndarray.sum)
    def sum(self, axis=None, dtype=None, out=None):
        ret = self.magnitude.sum(axis, dtype, None if out is None else out.magnitude)
        dim = self.dimensionality
        if out is None:
            return Quantity(ret, dim, copy=False)
        if not isinstance(out, Quantity):
            raise TypeError("out parameter must be a Quantity")
        out._dimensionality = dim
        return out

    @with_doc(np.nansum)
    def nansum(self, axis=None, dtype=None, out=None):
        import numpy as np
        return Quantity(
            np.nansum(self.magnitude, axis, dtype, out),
            self.dimensionality,
            copy=False
        )

    @with_doc(np.ndarray.fill)
    def fill(self, value):
        self.magnitude.fill(value)
        try:
            self._dimensionality = value.dimensionality
        except AttributeError:
            pass

    @with_doc(np.ndarray.put)
    def put(self, indicies, values, mode='raise', dtype='d'):
        """
        performs the equivalent of ndarray.put() but enforces units
        values - must be an Quantity with the same units as self
        """
        # The default of `dtype` is set to 'd' to ensure backwards
        # compatibility, arguably it's questionable design.
        if not isinstance(values, Quantity):
            values = Quantity(values)
        if values._dimensionality != self._dimensionality:
            values = values.rescale(self.units, dtype=dtype)
        self.magnitude.put(indicies, values, mode)

    # choose does not function correctly, and it is not clear
    # how it would function, so for now it will not be implemented

    @with_doc(np.ndarray.argsort)
    def argsort(self, axis=-1, kind='quick', order=None):
        return self.magnitude.argsort(axis, kind, order)

    @with_doc(np.ndarray.searchsorted)
    def searchsorted(self,values, side='left'):
        if not isinstance (values, Quantity):
            values = Quantity(values, copy=False)

        if values._dimensionality != self._dimensionality:
            raise ValueError("values does not have the same units as self")

        return self.magnitude.searchsorted(values.magnitude, side)

    @with_doc(np.ndarray.nonzero)
    def nonzero(self):
        return self.magnitude.nonzero()

    @with_doc(np.ndarray.max)
    def max(self, axis=None, out=None):
        ret = self.magnitude.max(axis, None if out is None else out.magnitude)
        dim = self.dimensionality
        if out is None:
            return Quantity(ret, dim, copy=False)
        if not isinstance(out, Quantity):
            raise TypeError("out parameter must be a Quantity")
        out._dimensionality = dim
        return out

    @with_doc(np.ndarray.argmax)
    def argmax(self, axis=None, out=None):
        return self.magnitude.argmax(axis, out)

    @with_doc(np.nanmax)
    def nanmax(self, axis=None, out=None):
        return Quantity(
            np.nanmax(self.magnitude),
            self.dimensionality,
            copy=False
        )

    @with_doc(np.ndarray.min)
    def min(self, axis=None, out=None):
        ret = self.magnitude.min(axis, None if out is None else out.magnitude)
        dim = self.dimensionality
        if out is None:
            return Quantity(ret, dim, copy=False)
        if not isinstance(out, Quantity):
            raise TypeError("out parameter must be a Quantity")
        out._dimensionality = dim
        return out

    @with_doc(np.nanmin)
    def nanmin(self, axis=None, out=None):
        return Quantity(
            np.nanmin(self.magnitude),
            self.dimensionality,
            copy=False
        )

    @with_doc(np.ndarray.argmin)
    def argmin(self, axis=None, out=None):
        return self.magnitude.argmin(axis, out)

    @with_doc(np.nanargmin)
    def nanargmin(self,axis=None, out=None):
        return np.nanargmin(self.magnitude)

    @with_doc(np.nanargmax)
    def nanargmax(self,axis=None, out=None):
        return np.nanargmax(self.magnitude)

    @with_doc(np.ndarray.ptp)
    def ptp(self, axis=None, out=None):
        ret = self.magnitude.ptp(axis, None if out is None else out.magnitude)
        dim = self.dimensionality
        if out is None:
            return Quantity(ret, dim, copy=False)
        if not isinstance(out, Quantity):
            raise TypeError("out parameter must be a Quantity")
        out._dimensionality = dim
        return out

    @with_doc(np.ndarray.clip)
    def clip(self, min=None, max=None, out=None):
        if min is None and max is None:
            raise ValueError("at least one of min or max must be set")
        else:
            if min is None: min = Quantity(-np.Inf, self._dimensionality)
            if max is None: max = Quantity(np.Inf, self._dimensionality)

        if self.dimensionality and not \
                (isinstance(min, Quantity) and isinstance(max, Quantity)):
            raise ValueError(
                "both min and max must be Quantities with compatible units"
            )

        clipped = self.magnitude.clip(
            min.rescale(self._dimensionality).magnitude,
            max.rescale(self._dimensionality).magnitude,
            out
        )
        dim = self.dimensionality
        if out is None:
            return Quantity(clipped, dim, copy=False)
        if not isinstance(out, Quantity):
            raise TypeError("out parameter must be a Quantity")
        out._dimensionality = dim
        return out

    @with_doc(np.ndarray.round)
    def round(self, decimals=0, out=None):
        ret = self.magnitude.round(decimals, None if out is None else out.magnitude)
        dim = self.dimensionality
        if out is None:
            return Quantity(ret, dim, copy=False)
        if not isinstance(out, Quantity):
            raise TypeError("out parameter must be a Quantity")
        out._dimensionality = dim
        return out

    @with_doc(np.ndarray.trace)
    def trace(self, offset=0, axis1=0, axis2=1, dtype=None, out=None):
        ret = self.magnitude.trace(offset, axis1, axis2, dtype, None if out is None else out.magnitude)
        dim = self.dimensionality
        if out is None:
            return Quantity(ret, dim, copy=False)
        if not isinstance(out, Quantity):
            raise TypeError("out parameter must be a Quantity")
        out._dimensionality = dim
        return out

    @with_doc(np.ndarray.squeeze)
    def squeeze(self, axis=None):
        return Quantity(
            self.magnitude.squeeze(axis),
            self.dimensionality,
            copy=False
        )

    @with_doc(np.ndarray.mean)
    def mean(self, axis=None, dtype=None, out=None):
        ret = self.magnitude.mean(axis, dtype, None if out is None else out.magnitude)
        dim = self.dimensionality
        if out is None:
            return Quantity(ret, dim, copy=False)
        if not isinstance(out, Quantity):
            raise TypeError("out parameter must be a Quantity")
        out._dimensionality = dim
        return out

    @with_doc(np.nanmean)
    def nanmean(self, axis=None, dtype=None, out=None):
        import numpy as np
        return Quantity(
            np.nanmean(self.magnitude, axis, dtype, out),
            self.dimensionality,
            copy=False)

    @with_doc(np.ndarray.var)
    def var(self, axis=None, dtype=None, out=None, ddof=0):
        ret = self.magnitude.var(axis, dtype, out, ddof)
        dim = self._dimensionality**2
        if out is None:
            return Quantity(ret, dim, copy=False)
        if not isinstance(out, Quantity):
            raise TypeError("out parameter must be a Quantity")
        out._dimensionality = dim
        return out

    @with_doc(np.ndarray.std)
    def std(self, axis=None, dtype=None, out=None, ddof=0):
        ret = self.magnitude.std(axis, dtype, out, ddof)
        dim = self.dimensionality
        if out is None:
            return Quantity(ret, dim, copy=False)
        if not isinstance(out, Quantity):
            raise TypeError("out parameter must be a Quantity")
        out._dimensionality = dim
        return out

    @with_doc(np.nanstd)
    def nanstd(self, axis=None, dtype=None, out=None, ddof=0):
        return Quantity(
            np.nanstd(self.magnitude, axis, dtype, out, ddof),
            self._dimensionality,
            copy=False
        )

    @with_doc(np.ndarray.prod)
    def prod(self, axis=None, dtype=None, out=None):
        if axis == None:
            power = self.size
        else:
            power = self.shape[axis]

        ret = self.magnitude.prod(axis, dtype, None if out is None else out.magnitude)
        dim = self._dimensionality**power
        if out is None:
            return Quantity(ret, dim, copy=False)
        if not isinstance(out, Quantity):
            raise TypeError("out parameter must be a Quantity")
        out._dimensionality = dim
        return out

    @with_doc(np.ndarray.cumsum)
    def cumsum(self, axis=None, dtype=None, out=None):
        ret = self.magnitude.cumsum(axis, dtype, None if out is None else out.magnitude)
        dim = self.dimensionality
        if out is None:
            return Quantity(ret, dim, copy=False)
        if not isinstance(out, Quantity):
            raise TypeError("out parameter must be a Quantity")
        out._dimensionality = dim
        return out

    @with_doc(np.ndarray.cumprod)
    def cumprod(self, axis=None, dtype=None, out=None):
        if self._dimensionality:
            # different array elements would have different dimensionality
            raise ValueError(
                "Quantity must be dimensionless, try using simplified"
            )

        ret = self.magnitude.cumprod(axis, dtype, out)
        dim = self.dimensionality
        if out is None:
            return Quantity(ret, dim, copy=False)
        if isinstance(out, Quantity):
            out._dimensionality = dim
        return out

    # list of unsupported functions: [choose]

    def __setstate__(self, state):
        ndarray_state = state[:-1]
        units = state[-1]
        np.ndarray.__setstate__(self, ndarray_state)
        self._dimensionality = units

    def __reduce__(self):
        """
        Return a tuple for pickling a Quantity.
        """
        reconstruct,reconstruct_args,state = super().__reduce__()
        state = state + (self._dimensionality,)
        return (_reconstruct_quantity,
                (self.__class__, np.ndarray, (0, ), 'b', ),
                state)

    def __deepcopy__(self, memo_dict):
        # constructor copies by default
        return Quantity(self.magnitude, self.dimensionality)


def _reconstruct_quantity(subtype, baseclass, baseshape, basetype,):
    """Internal function that builds a new MaskedArray from the
    information stored in a pickle.

    """
    _data = np.ndarray.__new__(baseclass, baseshape, basetype)
    return subtype.__new__(subtype, _data, dtype=basetype,)



"""
"""

from ..unitquantity import UnitQuantity
from .substance import mol
from .volume import L

M = molar = UnitQuantity(
    'molar',
    mol / L,
    symbol='M',
    aliases=['Molar']
)

mM = millimolar = UnitQuantity(
    'millimolar',
    molar / 1000,
    symbol='mM'
)

uM = micromolar = UnitQuantity(
    'micromolar',
    mM / 1000,
    symbol='uM',
    u_symbol='µM'
)



"""
"""

from ..unitquantity import UnitQuantity
from .mass import gram, kg, ounce, lb
from .length import cm, m, ft
from .time import s
from .acceleration import g_0


N = newton = UnitQuantity(
    'newton',
    kg*m/s**2,
    symbol='N',
    aliases=['newtons']
)
dyne = UnitQuantity(
    'dyne',
    gram*cm/s**2,
    symbol='dyn',
    aliases=['dynes']
)
pond = UnitQuantity(
    'pond',
    g_0*kg,
    symbol='p',
    aliases=['ponds']
)
kgf = force_kilogram = kilogram_force = UnitQuantity(
    'kilogram_force',
    kg*g_0,
    symbol='kgf',
    aliases=['force_kilogram']
)
ozf = force_ounce = ounce_force = UnitQuantity(
    'ounce_force',
    ounce*g_0,
    symbol='ozf',
    aliases=['force_ounce']
)
lbf = force_pound = pound_force = UnitQuantity(
    'pound_force',
    lb*g_0,
    symbol='lbf',
    aliases=['force_pound']
)
poundal = UnitQuantity(
    'poundal',
    lb*ft/s**2,
    symbol='pdl',
    aliases=['poundals']
)
gf = gram_force = force_gram = UnitQuantity(
    'gram_force',
    gram*g_0,
    symbol='gf',
    aliases=['force_gram']
)
force_ton = ton_force = UnitQuantity(
    'ton_force',
    2000*force_pound,
    aliases=['force_ton'])
kip = UnitQuantity(
    'kip', 1000*lbf
)

del UnitQuantity, gram, kg, cm, m, s, g_0



import numpy as np

from .. import units as pq
from .common import TestCase, unittest


class TestUmath(TestCase):

    @property
    def q(self):
        return [1,2,3,4] * pq.J

    def test_prod(self):
        self.assertQuantityEqual(np.prod(self.q), 24 * pq.J**4)

    def test_sum(self):
        self.assertQuantityEqual(np.sum(self.q), 10 * pq.J)

    def test_nansum(self):
        c = [1,2,3, np.NaN] * pq.m
        self.assertQuantityEqual(np.nansum(c), 6 * pq.m)

    def test_cumprod(self):
        self.assertRaises(ValueError, np.cumprod, self.q)

        q = [10, .1, 5, 50] * pq.dimensionless
        self.assertQuantityEqual(np.cumprod(q), [10, 1, 5, 250])

    def test_cumsum(self):
        self.assertQuantityEqual(np.cumsum(self.q), [1, 3, 6, 10] * pq.J)

    def test_diff(self):
        self.assertQuantityEqual(np.diff(self.q, 1), [1, 1, 1] * pq.J)

    def test_ediff1d(self):
        self.assertQuantityEqual(np.diff(self.q, 1), [1, 1, 1] * pq.J)

    def test_linspace(self):
        self.assertQuantityEqual(np.linspace(self.q[0], self.q[-1], 4), self.q)

    @unittest.expectedFailure
    def test_gradient(self):
        try:
            l = np.gradient([[1,1],[3,4]] * pq.J, 1 * pq.m)
            self.assertQuantityEqual(l[0], [[2., 3.], [2., 3.]] * pq.J/pq.m)
            self.assertQuantityEqual(l[1], [[0., 0.], [1., 1.]] * pq.J/pq.m)
        except ValueError as e:
            raise self.failureException(e)

    @unittest.expectedFailure
    def test_cross(self):
        a = [3,-3, 1] * pq.kPa
        b = [4, 9, 2] * pq.m**2
        self.assertQuantityEqual(np.cross(a,b), [-15,-2,39]*pq.kPa*pq.m**2)

    def test_trapz(self):
        self.assertQuantityEqual(np.trapz(self.q, dx = 1*pq.m), 7.5 * pq.J*pq.m)

    def test_sinh(self):
        q = [1, 2, 3, 4, 6] * pq.radian
        self.assertQuantityEqual(
            np.sinh(q),
            np.sinh(q.magnitude)
            )

    def test_arcsinh(self):
        q = [1, 2, 3, 4, 6] * pq.dimensionless
        self.assertQuantityEqual(
            np.arcsinh(q),
            np.arcsinh(q.magnitude) * pq.rad
            )

    def test_cosh(self):
        q = [1, 2, 3, 4, 6] * pq.radian
        self.assertQuantityEqual(
            np.cosh(q),
            np.cosh(q.magnitude) * pq.dimensionless
            )

    def test_arccosh(self):
        q = [1, 2, 3, 4, 6] * pq.dimensionless
        self.assertQuantityEqual(
            np.arccosh(q),
            np.arccosh(q.magnitude) * pq.rad
            )

    def test_tanh(self):
        q = [1, 2, 3, 4, 6] * pq.rad
        self.assertQuantityEqual(
            np.tanh(q),
            np.tanh(q.magnitude)
            )

    def test_arctanh(self):
        q = [.01, .5, .6, .8, .99] * pq.dimensionless
        self.assertQuantityEqual(
            np.arctanh(q),
            np.arctanh(q.magnitude) * pq.rad
            )

    def test_around(self):
        self.assertQuantityEqual(
            np.around([.5, 1.5, 2.5, 3.5, 4.5] * pq.J) ,
            [0., 2., 2., 4., 4.] * pq.J
            )

        self.assertQuantityEqual(
            np.around([1,2,3,11] * pq.J, decimals=1),
            [1, 2, 3, 11] * pq.J
            )

        self.assertQuantityEqual(
            np.around([1,2,3,11] * pq.J, decimals=-1),
            [0, 0, 0, 10] * pq.J
            )

    def test_round_(self):
        self.assertQuantityEqual(
            np.round_([.5, 1.5, 2.5, 3.5, 4.5] * pq.J),
            [0., 2., 2., 4., 4.] * pq.J
            )

        self.assertQuantityEqual(
            np.round_([1,2,3,11] * pq.J, decimals=1),
            [1, 2, 3, 11] * pq.J
            )

        self.assertQuantityEqual(
            np.round_([1,2,3,11] * pq.J, decimals=-1),
            [0, 0, 0, 10] * pq.J
            )

    def test_rint(self):
        a = [-4.1, -3.6, -2.5, 0.1, 2.5, 3.1, 3.9] * pq.m
        self.assertQuantityEqual(
            np.rint(a),
            [-4., -4., -2., 0., 2., 3., 4.]*pq.m
            )

    def test_floor(self):
        a = [-1.7, -1.5, -0.2, 0.2, 1.5, 1.7, 2.0] * pq.m
        self.assertQuantityEqual(
            np.floor(a),
            [-2., -2., -1., 0., 1., 1., 2.] * pq.m
            )

    def test_ceil(self):
        a = [-1.7, -1.5, -0.2, 0.2, 1.5, 1.7, 2.0] * pq.m
        self.assertQuantityEqual(
            np.ceil(a),
            [-1., -1., -0., 1., 2., 2., 2.] * pq.m
            )

    def test_fix(self):
        self.assertQuantityEqual(np.fix(3.14 * pq.degF), 3.0 * pq.degF)
        self.assertQuantityEqual(np.fix(3.0 * pq.degF), 3.0 * pq.degF)
        self.assertQuantityEqual(
            np.fix([2.1, 2.9, -2.1, -2.9] * pq.degF),
            [2., 2., -2., -2.] * pq.degF
            )

    def test_exp(self):
        self.assertQuantityEqual(np.exp(1*pq.dimensionless), np.e)
        self.assertRaises(ValueError, np.exp, 1*pq.m)

    def test_exp2(self):
        self.assertQuantityEqual(np.exp2(1*pq.dimensionless), 2.0)
        self.assertRaises(ValueError, np.exp2, 1*pq.m)

    def test_log(self):
        self.assertQuantityEqual(np.log(1*pq.dimensionless), 0)
        self.assertRaises(ValueError, np.log, 1*pq.m)

    def test_log10(self):
        self.assertQuantityEqual(np.log10(1*pq.dimensionless), 0)
        self.assertRaises(ValueError, np.log10, 1*pq.m)

    def test_log2(self):
        self.assertQuantityEqual(np.log2(1*pq.dimensionless), 0)
        self.assertRaises(ValueError, np.log2, 1*pq.m)

    def test_expm1(self):
        self.assertQuantityEqual(np.expm1(1*pq.dimensionless), np.e-1)
        self.assertRaises(ValueError, np.expm1, 1*pq.m)

    def test_log1p(self):
        self.assertQuantityEqual(np.log1p(0*pq.dimensionless), 0)
        self.assertRaises(ValueError, np.log1p, 1*pq.m)

    def test_sin(self):
        self.assertQuantityEqual(np.sin(np.pi/2*pq.radian), 1)
        self.assertRaises(ValueError, np.sin, 1*pq.m)

    def test_arcsin(self):
        self.assertQuantityEqual(
            np.arcsin(1*pq.dimensionless),
            np.pi/2 * pq.radian
            )
        self.assertRaises(ValueError, np.arcsin, 1*pq.m)

    def test_cos(self):
        self.assertQuantityEqual(np.cos(np.pi*pq.radians), -1)
        self.assertRaises(ValueError, np.cos, 1*pq.m)

    def test_arccos(self):
        self.assertQuantityEqual(np.arccos(1*pq.dimensionless), 0*pq.radian)
        self.assertRaises(ValueError, np.arccos, 1*pq.m)

    def test_tan(self):
        self.assertQuantityEqual(np.tan(0*pq.radian), 0)
        self.assertRaises(ValueError, np.tan, 1*pq.m)

    def test_arctan(self):
        self.assertQuantityEqual(np.arctan(0*pq.dimensionless), 0*pq.radian)
        self.assertRaises(ValueError, np.arctan, 1*pq.m)

    def test_arctan2(self):
        self.assertQuantityEqual(
            np.arctan2(0*pq.dimensionless, 0*pq.dimensionless),
            0
            )
        self.assertQuantityEqual(
            np.arctan2(3*pq.V, 3*pq.V),
            np.radians(45)*pq.dimensionless
            )
        # NumPy <1.21 raises ValueError
        # NumPy >=1.21 raises TypeError
        self.assertRaises((TypeError, ValueError), np.arctan2, (1*pq.m, 1*pq.m))

    def test_hypot(self):
        self.assertQuantityEqual(np.hypot(3 * pq.m, 4 * pq.m),  5 * pq.m)
        self.assertRaises(ValueError, np.hypot, 1*pq.m, 2*pq.J)

    def test_degrees(self):
        self.assertQuantityEqual(
            np.degrees(6 * pq.radians),
            (6 * pq.radians).rescale(pq.degree)
            )
        self.assertRaises(ValueError, np.degrees, 0*pq.degree)

    def test_radians(self):
        self.assertQuantityEqual(
            np.radians(6 * pq.degree),
            (6 * pq.degree).rescale(pq.radian)
            )
        self.assertRaises(ValueError, np.radians, 0*pq.radians)

    @unittest.expectedFailure
    def test_unwrap(self):
        self.assertQuantityEqual(np.unwrap([0,3*np.pi]*pq.radians), [0,np.pi])
        self.assertQuantityEqual(np.unwrap([0,540]*pq.deg), [0,180]*pq.deg)

    def test_equal(self):
        arr1 = (1, 1) * pq.m
        arr2 = (1.0, 1.0) * pq.m
        self.assertTrue(np.all(np.equal(arr1, arr2)))
        self.assertFalse(np.all(np.equal(arr1, arr2 * 2)))

    def test_not_equal(self):
        arr1 = (1, 1) * pq.m
        arr2 = (1.0, 1.0) * pq.m
        self.assertTrue(np.all(np.not_equal(arr1, arr2*2)))
        self.assertFalse(np.all(np.not_equal(arr1, arr2)))

    def test_less(self):
        arr1 = (1, 1) * pq.m
        arr2 = (1.0, 1.0) * pq.m
        self.assertTrue(np.all(np.less(arr1, arr2*2)))
        self.assertFalse(np.all(np.less(arr1*2, arr2)))

    def test_less_equal(self):
        arr1 = (1, 1) * pq.m
        arr2 = (1.0, 2.0) * pq.m
        self.assertTrue(np.all(np.less_equal(arr1, arr2)))
        self.assertFalse(np.all(np.less_equal(arr2, arr1)))

    def test_greater(self):
        arr1 = (1, 1) * pq.m
        arr2 = (1.0, 2.0) * pq.m
        self.assertTrue(np.all(np.greater(arr2*1.01, arr1)))
        self.assertFalse(np.all(np.greater(arr2, arr1)))

    def test_greater_equal(self):
        arr1 = (1, 1) * pq.m
        arr2 = (1.0, 2.0) * pq.m
        self.assertTrue(np.all(np.greater_equal(arr2, arr1)))
        self.assertFalse(np.all(np.greater_equal(arr2*0.99, arr1)))



"""
"""

from ._utils import _cd
from ..unitquantity import UnitConstant


R_d = deuteron_rms_charge_radius = UnitConstant(
    'deuteron_rms_charge_radius',
    _cd('deuteron rms charge radius'),
    symbol='R_d'
)
m_d = deuteron_mass = UnitConstant(
    'deuteron_mass',
    _cd('deuteron mass'),
    symbol='m_d'
)
g_d = deuteron_g_factor = UnitConstant(
    'deuteron_g_factor',
    _cd('deuteron g factor'),
    symbol='g_d'
)
mu_d = deuteron_magnetic_moment = UnitConstant(
    'deuteron_magnetic_moment',
    _cd('deuteron magnetic moment'),
    symbol='mu_d',
    u_symbol='μ_d'
)
deuteron_mass_energy_equivalent = UnitConstant(
    'deuteron_mass_energy_equivalent',
    _cd('deuteron mass energy equivalent'),
    symbol='(m_d*c**2)',
    u_symbol='(m_d·c²)'
)
deuteron_mass_energy_equivalent_in_MeV = UnitConstant(
    'deuteron_mass_energy_equivalent_in_MeV',
    _cd('deuteron mass energy equivalent in MeV')
)
deuteron_mass_in_u = UnitConstant(
    'deuteron_mass_in_u',
    _cd('deuteron mass in u')
)
deuteron_molar_mass = UnitConstant(
    'deuteron_molar_mass',
    _cd('deuteron molar mass'),
    symbol='M_d'
)
deuteron_electron_mass_ratio = UnitConstant(
    'deuteron_electron_mass_ratio',
    _cd('deuteron-electron mass ratio'),
    symbol='(m_d/m_e)',
    u_symbol='(m_d/mₑ)'
)
deuteron_proton_mass_ratio = UnitConstant(
    'deuteron_proton_mass_ratio',
    _cd('deuteron-proton mass ratio'),
    symbol='(m_d/m_n)'
)
deuteron_electron_magnetic_moment_ratio = UnitConstant(
    'deuteron_electron_magnetic_moment_ratio',
    _cd('deuteron-electron magnetic moment ratio'),
    symbol='(mu_d/mu_e)',
    u_symbol='(μ_d/μₑ)'
)
deuteron_magnetic_moment_to_Bohr_magneton_ratio = UnitConstant(
    'deuteron_magnetic_moment_to_Bohr_magneton_ratio',
    _cd('deuteron magnetic moment to Bohr magneton ratio'),
    symbol='(mu_d/mu_B)',
    u_symbol='(μ_d/μ_B)'
)
deuteron_magnetic_moment_to_nuclear_magneton_ratio = UnitConstant(
    'deuteron_magnetic_moment_to_nuclear_magneton_ratio',
    _cd('deuteron magnetic moment to nuclear magneton ratio'),
    symbol='(mu_d/mu_N)',
    u_symbol='(μ_d/μ_N)'
)
deuteron_neutron_magnetic_moment_ratio = UnitConstant(
    'deuteron_neutron_magnetic_moment_ratio',
    _cd('deuteron-neutron magnetic moment ratio'),
    symbol='(mu_d/mu_n)',
    u_symbol='(μ_d/μ_n)'
)
deuteron_proton_magnetic_moment_ratio = UnitConstant(
    'deuteron_proton_magnetic_moment_ratio',
    _cd('deuteron-proton magnetic moment ratio'),
    symbol='(mu_d/mu_p)',
    u_symbol='(μ_d/μ_p)'
)

del UnitConstant, _cd



import math as _math

from ..unitquantity import UnitConstant


pi = UnitConstant(
    'pi',
    _math.pi,
    u_symbol='π'
)
golden = golden_ratio = UnitConstant(
    'golden_ratio',
    (1 + _math.sqrt(5)) / 2,
    u_symbol='ϕ',
    aliases='golden'
)



"""
"""

from ._utils import _cd
from ..unitquantity import UnitConstant


amu = atomic_mass_constant = UnitConstant(
    'atomic_mass_constant',
    _cd('atomic mass constant'),
    symbol='m_u',
    u_symbol='mᵤ'
)
atomic_unit_of_1st_hyperpolarizability = UnitConstant(
    'atomic_unit_of_1st_hyperpolarizability',
    _cd('atomic unit of 1st hyperpolarizability'),
    symbol='(e**3*a_0**3/E_h**2)',
    u_symbol='(e³·a₀³/E_h²)'
)
atomic_unit_of_2nd_hyperpolarizability = UnitConstant(
    'atomic_unit_of_2nd_hyperpolarizability',
    _cd('atomic unit of 2nd hyperpolarizability'),
    symbol='(e**4*a_0**4/E_h**3)',
    u_symbol='(e⁴·a₀⁴/E_h³)'
)
hbar = atomic_unit_of_action = UnitConstant(
    'atomic_unit_of_action',
    _cd('atomic unit of action'),
    symbol='hbar',
    u_symbol='ħ'
)
atomic_unit_of_charge = UnitConstant(
    'atomic_unit_of_charge',
    _cd('atomic unit of charge'),
    symbol='e'
)
atomic_unit_of_charge_density = UnitConstant(
    'atomic_unit_of_charge_density',
    _cd('atomic unit of charge density'),
    symbol='(e/a_0**3)',
    u_symbol='(e/a₀³)'
)
atomic_unit_of_current = UnitConstant(
    'atomic_unit_of_current',
    _cd('atomic unit of current'),
    symbol='(e*E_h/hbar)',
    u_symbol='(e·E_h/ħ)'
)
atomic_unit_of_electric_dipole_moment = UnitConstant(
    'atomic_unit_of_electric_dipole_moment',
    _cd('atomic unit of electric dipole moment'),
    symbol='(e*a_0)',
    u_symbol='(e·a₀)'
)
atomic_unit_of_electric_field = UnitConstant(
    'atomic_unit_of_electric_field',
    _cd('atomic unit of electric field'),
    symbol='(E_h/(e*a_0))',
    u_symbol='(E_h/(e·a₀))'
)
atomic_unit_of_electric_field_gradient = UnitConstant(
    'atomic_unit_of_electric_field_gradient',
    _cd('atomic unit of electric field gradient'),
    symbol='(E_h/(e*a_0**2))',
    u_symbol='(E_h/(e·a₀²))'
)
atomic_unit_of_electric_polarizability = UnitConstant(
    'atomic_unit_of_electric_polarizability',
    _cd('atomic unit of electric polarizability'),
    symbol='(e**2*a_0**2/E_h)',
    u_symbol='(e²·a₀²/E_h)'
)
atomic_unit_of_electric_potential = UnitConstant(
    'atomic_unit_of_electric_potential',
    _cd('atomic unit of electric potential'),
    symbol='(E_h/e)'
)
atomic_unit_of_electric_quadrupole_moment = UnitConstant(
    'atomic_unit_of_electric_quadrupole_moment',
    _cd('atomic unit of electric quadrupole moment'),
    symbol='(e*a_0**2)',
    u_symbol='(e·a₀²)'
)
atomic_unit_of_energy = UnitConstant(
    'atomic_unit_of_energy',
    _cd('atomic unit of energy'),
)
atomic_unit_of_force = UnitConstant(
    'atomic_unit_of_force',
    _cd('atomic unit of force'),
    symbol='(E_h/a_0)',
    u_symbol='(E_h/a₀)'
)
a_0 = atomic_unit_of_length = UnitConstant(
    'atomic_unit_of_length',
    _cd('atomic unit of length'),
    symbol='a_0',
    u_symbol='a₀'
)
Bohr_radius = UnitConstant(
    'Bohr_radius',
    _cd('Bohr radius'),
    symbol='a_0',
    u_symbol='a₀'
)
atomic_unit_of_magnetic_dipole_moment = UnitConstant(
    'atomic_unit_of_magnetic_dipole_moment',
    _cd('atomic unit of magnetic dipole moment'),
    symbol='(hbar*e/m_e)',
    u_symbol='(ħ·e/mₑ)'
)
atomic_unit_of_magnetic_flux_density = UnitConstant(
    'atomic_unit_of_magnetic_flux_density',
    _cd('atomic unit of magnetic flux density'),
    symbol='(hbar*e/a_0**2)',
    u_symbol='(ħ·e/a₀²)'
)
atomic_unit_of_magnetizability = UnitConstant(
    'atomic_unit_of_magnetizability',
    _cd('atomic unit of magnetizability'),
    symbol='(e**2a_0**2/m_e)',
    u_symbol='(e²·a₀²/mₑ)'
)
m_e = atomic_unit_of_mass = UnitConstant(
    'atomic_unit_of_mass',
    _cd('atomic unit of mass'),
    symbol='m_e',
    u_symbol='mₑ'
)
atomic_unit_of_momentum = UnitConstant(
    'atomic_unit_of_momentum',
    _cd('atomic unit of momentum'),
    symbol='(hbar/a_0)',
    u_symbol='(ħ/a₀)'
)
atomic_unit_of_permittivity = UnitConstant(
    'atomic_unit_of_permittivity',
    _cd('atomic unit of permittivity'),
    symbol='(e**2/(a_0*E_h))',
    u_symbol='(e²/(a₀·E_h))'
)
atomic_unit_of_time = UnitConstant(
    'atomic_unit_of_time',
    _cd('atomic unit of time'),
    symbol='(hbar/E_h)',
    u_symbol='(ħ/E_h)'
)
atomic_unit_of_velocity = UnitConstant(
    'atomic_unit_of_velocity',
    _cd('atomic unit of velocity'),
    symbol='(a_0*E_h/hbar)',
    u_symbol='(a₀·E_h/ħ)'
)
E_h = Hartree_energy = UnitConstant(
    'Hartree_energy',
    _cd('Hartree energy'),
    symbol='E_h'
)
u = unified_atomic_mass_unit = UnitConstant(
    'unified_atomic_mass_unit',
    _cd('unified atomic mass unit'),
    symbol='u'
)
molar_mass_of_carbon_12 = UnitConstant(
    'molar_mass_of_carbon_12',
    _cd('molar mass of carbon-12'),
    symbol='M_C12',
    u_symbol='M_¹²C'
)

atomic_mass_constant_energy_equivalent = UnitConstant(
    'atomic_mass_constant_energy_equivalent',
    _cd('atomic mass constant energy equivalent'),
    symbol='(m_u*c**2)',
    u_symbol='(mᵤ·c²)'
)
atomic_mass_constant_energy_equivalent_in_MeV = UnitConstant(
    'atomic_mass_constant_energy_equivalent_in_MeV',
    _cd('atomic mass constant energy equivalent in MeV')
)
Hartree_energy_in_eV = UnitConstant(
    'Hartree_energy_in_eV',
    _cd('Hartree energy in eV')
)

atomic_mass_unit_electron_volt_relationship = UnitConstant(
    'atomic_mass_unit_electron_volt_relationship',
    _cd('atomic mass unit-electron volt relationship'),
)
atomic_mass_unit_hartree_relationship = UnitConstant(
    'atomic_mass_unit_hartree_relationship',
    _cd('atomic mass unit-hartree relationship')
)
atomic_mass_unit_hertz_relationship = UnitConstant(
    'atomic_mass_unit_hertz_relationship',
    _cd('atomic mass unit-hertz relationship')
)
atomic_mass_unit_inverse_meter_relationship = UnitConstant(
    'atomic_mass_unit_inverse_meter_relationship',
    _cd('atomic mass unit-inverse meter relationship')
)
atomic_mass_unit_joule_relationship = UnitConstant(
    'atomic_mass_unit_joule_relationship',
    _cd('atomic mass unit-joule relationship'),
    symbol='(u*c**2)',
    u_symbol='(u·c²)'
)
atomic_mass_unit_kelvin_relationship = UnitConstant(
    'atomic_mass_unit_kelvin_relationship',
    _cd('atomic mass unit-kelvin relationship')
)
atomic_mass_unit_kilogram_relationship = UnitConstant(
    'atomic_mass_unit_kilogram_relationship',
    _cd('atomic mass unit-kilogram relationship')
)

hartree_atomic_mass_unit_relationship = UnitConstant(
    'hartree_atomic_mass_unit_relationship',
    _cd('hartree-atomic mass unit relationship')
)
hartree_electron_volt_relationship = UnitConstant(
    'hartree_electron_volt_relationship',
    _cd('hartree-electron volt relationship')
)
hartree_hertz_relationship = UnitConstant(
    'hartree_hertz_relationship',
    _cd('hartree-hertz relationship')
)
hartree_inverse_meter_relationship = UnitConstant(
    'hartree_inverse_meter_relationship',
    _cd('hartree-inverse meter relationship')
)
hartree_joule_relationship = UnitConstant(
    'hartree_joule_relationship',
    _cd('hartree-joule relationship')
)
hartree_kelvin_relationship = UnitConstant(
    'hartree_kelvin_relationship',
    _cd('hartree-kelvin relationship')
)
hartree_kilogram_relationship = UnitConstant(
    'hartree_kilogram_relationship',
    _cd('hartree-kilogram relationship')
)
hertz_atomic_mass_unit_relationship = UnitConstant(
    'hertz_atomic_mass_unit_relationship',
    _cd('hertz-atomic mass unit relationship')
)
hertz_hartree_relationship = UnitConstant(
    'hertz_hartree_relationship',
    _cd('hertz-hartree relationship')
)
inverse_meter_atomic_mass_unit_relationship = UnitConstant(
    'inverse_meter_atomic_mass_unit_relationship',
    _cd('inverse meter-atomic mass unit relationship')
)
inverse_meter_hartree_relationship = UnitConstant(
    'inverse_meter_hartree_relationship',
    _cd('inverse meter-hartree relationship')
)
joule_atomic_mass_unit_relationship = UnitConstant(
    'joule_atomic_mass_unit_relationship',
    _cd('joule-atomic mass unit relationship')
)
joule_hartree_relationship = UnitConstant(
    'joule_hartree_relationship',
    _cd('joule-hartree relationship')
)
kelvin_atomic_mass_unit_relationship = UnitConstant(
    'kelvin_atomic_mass_unit_relationship',
    _cd('kelvin-atomic mass unit relationship')
)
kelvin_hartree_relationship = UnitConstant(
    'kelvin_hartree_relationship',
    _cd('kelvin-hartree relationship')
)
kilogram_atomic_mass_unit_relationship = UnitConstant(
    'kilogram_atomic_mass_unit_relationship',
    _cd('kilogram-atomic mass unit relationship')
)
kilogram_hartree_relationship = UnitConstant(
    'kilogram_hartree_relationship',
    _cd('kilogram-hartree relationship')
)

del UnitConstant, _cd



"""
"""

import numpy as np

from . import markup
from .quantity import Quantity, scale_other_units
from .registry import unit_registry
from .decorators import with_doc

class UncertainQuantity(Quantity):

    # TODO: what is an appropriate value?
    __array_priority__ = 22

    def __new__(cls, data, units='', uncertainty=None, dtype='d', copy=True):
        ret = Quantity.__new__(cls, data, units, dtype, copy)
        # _uncertainty initialized to be dimensionless by __array_finalize__:
        ret._uncertainty._dimensionality = ret._dimensionality

        if uncertainty is not None:
            ret.uncertainty = Quantity(uncertainty, ret._dimensionality)
        elif isinstance(data, UncertainQuantity):
            if copy or ret._dimensionality != uncertainty._dimensionality:
                uncertainty = data.uncertainty.rescale(ret.units)
            ret.uncertainty = uncertainty

        return ret

    @Quantity.units.setter
    def units(self, units):
        super()._set_units(units)
        self.uncertainty.units = self._dimensionality

    @property
    def _reference(self):
        ret = super()._reference.view(UncertainQuantity)
        ret.uncertainty = self.uncertainty._reference
        return ret

    @property
    def simplified(self):
        ret = super().simplified.view(UncertainQuantity)
        ret.uncertainty = self.uncertainty.simplified
        return ret

    @property
    def uncertainty(self):
        return self._uncertainty
    @uncertainty.setter
    def uncertainty(self, uncertainty):
        if not isinstance(uncertainty, Quantity):
            uncertainty = Quantity(uncertainty, copy=False)
        try:
            assert self.shape == uncertainty.shape
        except AssertionError:
            raise ValueError('data and uncertainty must have identical shape')
        if uncertainty._dimensionality != self._dimensionality:
            uncertainty = uncertainty.rescale(self._dimensionality)
        self._uncertainty = uncertainty

    @property
    def relative_uncertainty(self):
        return self.uncertainty.magnitude/self.magnitude

    @with_doc(Quantity.rescale, use_header=False)
    def rescale(self, units, dtype=None):
        cls = UncertainQuantity
        ret = super(cls, self).rescale(units, dtype=dtype).view(cls)
        ret.uncertainty = self.uncertainty.rescale(units, dtype=dtype)
        return ret

    def __array_finalize__(self, obj):
        Quantity.__array_finalize__(self, obj)
        self._uncertainty = getattr(
            obj,
            'uncertainty',
            Quantity(
                np.zeros(self.shape, self.dtype),
                self._dimensionality,
                copy=False
            )
        )

    @with_doc(Quantity.__add__, use_header=False)
    @scale_other_units
    def __add__(self, other):
        res = super().__add__(other)
        u = (self.uncertainty**2+other.uncertainty**2)**0.5
        return UncertainQuantity(res, uncertainty=u, copy=False)

    @with_doc(Quantity.__radd__, use_header=False)
    @scale_other_units
    def __radd__(self, other):
        return self.__add__(other)

    @with_doc(Quantity.__sub__, use_header=False)
    @scale_other_units
    def __sub__(self, other):
        res = super().__sub__(other)
        u = (self.uncertainty**2+other.uncertainty**2)**0.5
        return UncertainQuantity(res, uncertainty=u, copy=False)

    @with_doc(Quantity.__rsub__, use_header=False)
    @scale_other_units
    def __rsub__(self, other):
        if not isinstance(other, UncertainQuantity):
            other = UncertainQuantity(other, copy=False)

        return UncertainQuantity.__sub__(other, self)

    @with_doc(Quantity.__mul__, use_header=False)
    def __mul__(self, other):
        res = super().__mul__(other)
        try:
            sru = self.relative_uncertainty
            oru = other.relative_uncertainty
            ru = (sru**2+oru**2)**0.5
            u = res.view(Quantity) * ru
        except AttributeError:
            other = np.array(other, copy=False, subok=True)
            u = (self.uncertainty**2*other**2)**0.5

        res._uncertainty = u
        return res

    @with_doc(Quantity.__rmul__, use_header=False)
    def __rmul__(self, other):
        return self.__mul__(other)

    def __neg__(self):
        return self*-1

    @with_doc(Quantity.__truediv__, use_header=False)
    def __truediv__(self, other):
        res = super().__truediv__(other)
        try:
            sru = self.relative_uncertainty
            oru = other.relative_uncertainty
            ru = (sru**2+oru**2)**0.5
            u = res.view(Quantity) * ru
        except AttributeError:
            other = np.array(other, copy=False, subok=True)
            u = (self.uncertainty**2/other**2)**0.5

        res._uncertainty = u
        return res

    @with_doc(Quantity.__rtruediv__, use_header=False)
    def __rtruediv__(self, other):
        temp = UncertainQuantity(
            1/self.magnitude, self.dimensionality**-1,
            self.relative_uncertainty/self.magnitude, copy=False
        )
        return other * temp

    @with_doc(Quantity.__pow__, use_header=False)
    def __pow__(self, other):
        res = super().__pow__(other)
        res.uncertainty = res.view(Quantity) * other * self.relative_uncertainty
        return res

    @with_doc(Quantity.__getitem__, use_header=False)
    def __getitem__(self, key):
        return UncertainQuantity(
            self.magnitude[key],
            self._dimensionality,
            self.uncertainty[key],
            copy=False
        )

    @with_doc(Quantity.__repr__, use_header=False)
    def __repr__(self):
        return '%s(%s, %s, %s)'%(
            self.__class__.__name__,
            repr(self.magnitude),
            self.dimensionality.string,
            repr(self.uncertainty.magnitude)
        )

    @with_doc(Quantity.__str__, use_header=False)
    def __str__(self):
        if markup.config.use_unicode:
            dims = self.dimensionality.unicode
        else:
            dims = self.dimensionality.string
        s = '%s %s\n+/-%s (1 sigma)'%(
            str(self.magnitude),
            dims,
            str(self.uncertainty)
        )
        if markup.config.use_unicode:
            return s.replace('+/-', '±').replace(' sigma', 'σ')
        return s

    @with_doc(np.ndarray.sum)
    def sum(self, axis=None, dtype=None, out=None):
        return UncertainQuantity(
            self.magnitude.sum(axis, dtype, out),
            self.dimensionality,
            (np.sum(self.uncertainty.magnitude**2, axis))**0.5,
            copy=False
        )

    @with_doc(np.nansum)
    def nansum(self, axis=None, dtype=None, out=None):
        return UncertainQuantity(
            np.nansum(self.magnitude, axis, dtype, out),
            self.dimensionality,
            (np.nansum(self.uncertainty.magnitude**2, axis))**0.5,
            copy=False
        )

    @with_doc(np.ndarray.mean)
    def mean(self, axis=None, dtype=None, out=None):
        return UncertainQuantity(
            self.magnitude.mean(axis, dtype, out),
            self.dimensionality,
            ((1.0/np.size(self,axis))**2 * np.sum(self.uncertainty.magnitude**2, axis))**0.5,
            copy=False)

    @with_doc(np.nanmean)
    def nanmean(self, axis=None, dtype=None, out=None):
        size = np.sum(~np.isnan(self),axis)
        return UncertainQuantity(
            np.nanmean(self.magnitude, axis, dtype, out),
            self.dimensionality,
            ((1.0/size)**2 * np.nansum(np.nan_to_num(self.uncertainty.magnitude)**2, axis))**0.5,
            copy=False)

    @with_doc(np.sqrt)
    def sqrt(self, out=None):
        return self**0.5

    @with_doc(np.ndarray.max)
    def max(self, axis=None, out=None):
        idx = np.unravel_index(np.argmax(self.magnitude), self.shape)
        return self[idx]

    @with_doc(np.nanmax)
    def nanmax(self, axis=None, out=None):
        idx = np.unravel_index(np.nanargmax(self.magnitude), self.shape)
        return self[idx]

    @with_doc(np.ndarray.min)
    def min(self, axis=None, out=None):
        idx = np.unravel_index(np.argmin(self.magnitude), self.shape)
        return self[idx]

    @with_doc(np.nanmin)
    def nanmin(self, axis=None, out=None):
        idx = np.unravel_index(np.nanargmin(self.magnitude), self.shape)
        return self[idx]

    @with_doc(np.ndarray.argmin)
    def argmin(self,axis=None, out=None):
        return self.magnitude.argmin()

    @with_doc(np.ndarray.argmax)
    def argmax(self,axis=None, out=None):
        return self.magnitude.argmax()

    @with_doc(np.nanargmin)
    def nanargmin(self,axis=None, out=None):
        return np.nanargmin(self.magnitude)

    @with_doc(np.nanargmax)
    def nanargmax(self,axis=None, out=None):
        return np.nanargmax(self.magnitude)

    def __setstate__(self, state):
        ndarray_state = state[:-2]
        units, sigma = state[-2:]
        np.ndarray.__setstate__(self, ndarray_state)
        self._dimensionality = units
        self._uncertainty = sigma

    def __reduce__(self):
        """
        Return a tuple for pickling a Quantity.
        """
        reconstruct, reconstruct_args, state = super().__reduce__()
        state = state + (self._uncertainty,)
        return reconstruct, reconstruct_args, state



"""
"""

from ..unitquantity import UnitSubstance

mol = mole = UnitSubstance(
    'mole',
    symbol='mol'
)
mmol = UnitSubstance(
    'millimole',
    mol/1000,
    symbol='mmol'
)
umol = UnitSubstance(
    'micromole',
    mmol/1000,
    symbol='umol',
    u_symbol='µmol'
)



import pickle
import copy

from .. import units as pq
from ..quantity import Quantity
from ..uncertainquantity import UncertainQuantity
from .. import constants
from .common import TestCase


class TestPersistence(TestCase):

    def test_unitquantity_persistence(self):
        x = pq.m
        y = pickle.loads(pickle.dumps(x))
        self.assertQuantityEqual(x, y)

        x = pq.CompoundUnit("pc/cm**3")
        y = pickle.loads(pickle.dumps(x))
        self.assertQuantityEqual(x, y)

    def test_quantity_persistence(self):
        x = 20*pq.m
        y = pickle.loads(pickle.dumps(x))
        self.assertQuantityEqual(x, y)

    def test_uncertainquantity_persistence(self):
        x = UncertainQuantity(20, 'm', 0.2)
        y = pickle.loads(pickle.dumps(x))
        self.assertQuantityEqual(x, y)

    def test_unitconstant_persistence(self):
        x = constants.m_e
        y = pickle.loads(pickle.dumps(x))
        self.assertQuantityEqual(x, y)

    def test_quantity_object_dtype(self):
        # Regression test for github issue #113
        x = Quantity(1,dtype=object)
        y = pickle.loads(pickle.dumps(x))
        self.assertQuantityEqual(x, y)

    def test_uncertainquantity_object_dtype(self):
        # Regression test for github issue #113
        x = UncertainQuantity(20, 'm', 0.2, dtype=object)
        y = pickle.loads(pickle.dumps(x))
        self.assertQuantityEqual(x, y)

    def test_backward_compat(self):
        """ A few pickles collected before fixing #113 just to make sure we remain backwards compatible. """
        orig = [
            pq.m,
            20*pq.m,
            UncertainQuantity(20, 'm', 0.2),
            constants.m_e,
        ]
        data = [
            # generated with protocol=2
            b'\x80\x02cquantities.unitquantity\nUnitLength\nq\x00(X\x05\x00\x00\x00meterq\x01NX\x01\x00\x00\x00mq\x02N]q\x03(X\x06\x00\x00\x00metersq\x04X\x05\x00\x00\x00metreq\x05X\x06\x00\x00\x00metresq\x06eNtq\x07Rq\x08K\x01K\x02K\x02\x86q\t\x86q\nb.',
            b'\x80\x02cquantities.quantity\n_reconstruct_quantity\nq\x00(cquantities.quantity\nQuantity\nq\x01cnumpy\nndarray\nq\x02K\x00\x85q\x03X\x01\x00\x00\x00bq\x04tq\x05Rq\x06(K\x01)cnumpy\ndtype\nq\x07X\x02\x00\x00\x00f8q\x08K\x00K\x01\x87q\tRq\n(K\x03X\x01\x00\x00\x00<q\x0bNNNJ\xff\xff\xff\xffJ\xff\xff\xff\xffK\x00tq\x0cb\x89c_codecs\nencode\nq\rX\x08\x00\x00\x00\x00\x00\x00\x00\x00\x004@q\x0eX\x06\x00\x00\x00latin1q\x0f\x86q\x10Rq\x11cquantities.dimensionality\nDimensionality\nq\x12)\x81q\x13cquantities.unitquantity\nUnitLength\nq\x14(X\x05\x00\x00\x00meterq\x15NX\x01\x00\x00\x00mq\x16N]q\x17(X\x06\x00\x00\x00metersq\x18X\x05\x00\x00\x00metreq\x19X\x06\x00\x00\x00metresq\x1aeNtq\x1bRq\x1cK\x01K\x02K\x02\x86q\x1d\x86q\x1ebK\x01stq\x1fb.',
            b'\x80\x02cquantities.quantity\n_reconstruct_quantity\nq\x00(cquantities.uncertainquantity\nUncertainQuantity\nq\x01cnumpy\nndarray\nq\x02K\x00\x85q\x03X\x01\x00\x00\x00bq\x04tq\x05Rq\x06(K\x01)cnumpy\ndtype\nq\x07X\x02\x00\x00\x00f8q\x08K\x00K\x01\x87q\tRq\n(K\x03X\x01\x00\x00\x00<q\x0bNNNJ\xff\xff\xff\xffJ\xff\xff\xff\xffK\x00tq\x0cb\x89c_codecs\nencode\nq\rX\x08\x00\x00\x00\x00\x00\x00\x00\x00\x004@q\x0eX\x06\x00\x00\x00latin1q\x0f\x86q\x10Rq\x11cquantities.dimensionality\nDimensionality\nq\x12)\x81q\x13cquantities.unitquantity\nUnitLength\nq\x14(X\x05\x00\x00\x00meterq\x15NX\x01\x00\x00\x00mq\x16N]q\x17(X\x06\x00\x00\x00metersq\x18X\x05\x00\x00\x00metreq\x19X\x06\x00\x00\x00metresq\x1aeNtq\x1bRq\x1cK\x01K\x02K\x02\x86q\x1d\x86q\x1ebK\x01sh\x00(cquantities.quantity\nQuantity\nq\x1fh\x02h\x03h\x04tq Rq!(K\x01)h\n\x89h\rX\x0f\x00\x00\x00\xc2\x9a\xc2\x99\xc2\x99\xc2\x99\xc2\x99\xc2\x99\xc3\x89?q"h\x0f\x86q#Rq$h\x12)\x81q%h\x1cK\x01stq&btq\'b.',
            b'\x80\x02cquantities.unitquantity\nUnitConstant\nq\x00(X\r\x00\x00\x00electron_massq\x01cquantities.quantity\n_reconstruct_quantity\nq\x02(cquantities.quantity\nQuantity\nq\x03cnumpy\nndarray\nq\x04K\x00\x85q\x05X\x01\x00\x00\x00bq\x06tq\x07Rq\x08(K\x01)cnumpy\ndtype\nq\tX\x02\x00\x00\x00f8q\nK\x00K\x01\x87q\x0bRq\x0c(K\x03X\x01\x00\x00\x00<q\rNNNJ\xff\xff\xff\xffJ\xff\xff\xff\xffK\x00tq\x0eb\x89c_codecs\nencode\nq\x0fX\x0c\x00\x00\x00N?\xc3\xab\xc2\x93\xc3\x9cy\xc2\xb29q\x10X\x06\x00\x00\x00latin1q\x11\x86q\x12Rq\x13cquantities.dimensionality\nDimensionality\nq\x14)\x81q\x15cquantities.unitquantity\nUnitMass\nq\x16(X\x08\x00\x00\x00kilogramq\x17NX\x02\x00\x00\x00kgq\x18N]q\x19X\t\x00\x00\x00kilogramsq\x1aaNtq\x1bRq\x1cK\x01K\x01K?\x86q\x1d\x86q\x1ebK\x01stq\x1fbX\x03\x00\x00\x00m_eq X\x04\x00\x00\x00m\xe2\x82\x91q!]q"Ntq#Rq$K\x01K\x00M!\x01\x86q%\x86q&b.',
        ]
        for x,d in zip(orig,data):
            y = pickle.loads(d)
            self.assertQuantityEqual(x, y)

    def test_copy_quantity(self):
        for dtype in [float,object]:
            x = (20*pq.m).astype(dtype)
            y = copy.copy(x)
            self.assertQuantityEqual(x, y)

    def test_copy_uncertainquantity(self):
        for dtype in [float, object]:
            x = UncertainQuantity(20, 'm', 0.2).astype(dtype)
            y = copy.copy(x)
            self.assertQuantityEqual(x, y)


from .. import units as pq
from .common import TestCase


class TestFormatting(TestCase):

    @staticmethod
    def _check(quantity, formatted):
        assert str(quantity) == formatted
        assert f'{quantity}' == formatted
        assert f'{quantity!s}' == formatted

    def test_str_format_scalar(self):
        self._check(1*pq.J, '1.0 J')

    def test_str_format_non_scalar(self):
        self._check([1, 2]*pq.J, '[1. 2.] J')



"""
"""

from ..unitquantity import UnitQuantity
from .time import s
from .mass import kg
from .energy import J
from .electromagnetism import coulomb


Bq = becquerel = UnitQuantity(
    'becquerel',
    1/s,
    symbol='Bq',
    aliases=['becquerels']
)
Ci = curie = UnitQuantity(
    'curie',
    3.7e10*becquerel,
    symbol='Ci',
    aliases=['curies']
)
rd = rutherford = UnitQuantity(
    'rutherford',
    1e6*Bq,
    symbol='Rd',
    aliases=['rutherfords'],
    doc='this unit is obsolete, in favor of 1e6 Bq'
)
Gy = gray = Sv = sievert = UnitQuantity(
    'gray',
    J/kg,
    symbol='Gy',
    aliases=['grays', 'Sv', 'sievert', 'sieverts']
)
rem = UnitQuantity(
    'rem',
    1e-2*sievert,
    aliases=['rems']
)
rads = UnitQuantity(
    'rads',
    1e-2*gray,
    doc='''
    rad is commonly used symbol for radian. 
    rads unit of radiation is deprecated.
    '''
)
R = roentgen = UnitQuantity(
    'roentgen',
    2.58e-4*coulomb/kg,
    symbol='R',
    aliases=['roentgens']
)

del UnitQuantity, s, kg, J, coulomb



"""
"""

from ..unitquantity import UnitQuantity, UnitMass
from .length import m

kg = kilogram =  UnitMass(
    'kilogram',
    symbol='kg',
    aliases=['kilograms']
)
g = gram = UnitMass(
    'gram',
    kg/1000,
    symbol='g',
    aliases=['grams']
)
mg = milligram = UnitMass(
    'milligram',
    gram/1000,
    symbol='mg',
    aliases=['milligrams']
)
oz = ounce = avoirdupois_ounce = UnitMass(
    'ounce',
    28.349523125*g,
    symbol='oz',
    aliases=['ounces','avoirdupois_ounce', 'avoirdupois_ounces'],
    doc='exact'
)
lb = pound = avoirdupois_pound = UnitMass(
    'pound',
    0.45359237*kg,
    symbol='lb',
    aliases=['pounds', 'avoirdupois_pound', 'avoirdupois_pounds'],
    doc='exact'
)
st = stone = UnitMass(
    'stone',
    14*lb,
    symbol='st',
    doc='As defined in the UK, 1 stone = 14 avoirdupois pounds'
)

carat = UnitMass(
    'carat',
    200*mg,
    aliases=['carats']
)
gr = grain = UnitMass(
    'grain',
    64.79891*mg,
    symbol='gr',
    aliases=['grains']
)
long_hundredweight = UnitMass(
    'long_hundredweight',
    112*lb,
    aliases=['long_hundredweights']
)
short_hundredweight = UnitMass(
    'short_hundredweight',
    100*lb,
    aliases=['short_hundredweights']
) # cwt is used for both short and long hundredweight, so we wont use it
t = metric_ton = tonne = UnitMass(
    'tonne',
    1000*kg,
    symbol='t',
    aliases=['tonnes']
)
dwt = pennyweight = UnitMass(
    'pennyweight',
    24*gr,
    symbol='dwt',
    aliases=['pennyweights']
)
slug = slugs = UnitMass(
    'slug',
    14.59390*kg,
    aliases=['slugs']
)
toz = troy_ounce = apounce = apothecary_ounce = UnitMass(
    'troy_ounce',
    480*gr,
    symbol='toz',
    u_symbol='℥',
    aliases=[
        'apounce', 'apounces', 'apothecary_ounce', 'apothecary_ounces',
        'troy_ounces'
    ]
)
troy_pound = appound = apothecary_pound = UnitMass(
    'troy_pound',
    12*toz,
    symbol='tlb',
    u_symbol='℔',
    aliases=[
        'troy_pounds', 'appound', 'appounds', 'apothecary_pound',
        'apothecary_pounds'
    ]
)
u = amu = atomic_mass_unit = dalton = Da = UnitMass(
    'atomic_mass_unit',
    1.660538782e-27*kg,
    symbol='u',
    aliases=['amu', 'Da', 'dalton'],
    doc='relative uncertainty = 5e-8'
)
scruple = UnitMass(
    'scruple',
    20*gr,
    u_symbol='℈',
    aliases=['scruples']
)
dr = dram = UnitMass(
    'dram',
    oz/16,
    symbol='dr',
    aliases=['drams'],
    doc='avoirdupois dram'
)
drachm = apdram = UnitMass(
    'drachm',
    60*gr,
    u_symbol='  ',
    aliases=['drachms', 'apdram', 'apdrams'],
    doc='also known as the apothecary dram'
)

bag = UnitMass(
    'bag',
    94*lb,
    aliases=['bags']
)

ton = short_ton = UnitMass(
    'short_ton',
    2000*lb,
    aliases=['short_tons']
)
long_ton = UnitMass(
    'long_ton', 2240*lb,
    aliases=['long_tons']
) # both long and short tons are referred to as "ton" so we wont use it

############################################################
##                 Mass per unit length                   ##
############################################################

denier = UnitQuantity(
    'denier',
    g/(9000*m),
    aliases=['deniers']
)
tex = UnitQuantity(
    'tex',
    g/(1000*m),
    aliases=['texs']
)
dtex = UnitQuantity(
    'dtex',
    g/(10000*m),
    aliases=['dtexs']
)

del UnitQuantity, m



from .. import units as pq
from .common import TestCase

class TestUnits(TestCase):

    def test_compound_units(self):
        pc_per_cc = pq.CompoundUnit("pc/cm**3")
        self.assertEqual(str(pc_per_cc.dimensionality), "(pc/cm**3)")
        self.assertEqual(str(pc_per_cc), "1 (pc/cm**3)")

        temp = pc_per_cc * pq.CompoundUnit('m/m**3')
        self.assertEqual(str(temp.dimensionality), "(pc/cm**3)*(m/m**3)")
        self.assertEqual(str(temp), "1.0 (pc/cm**3)*(m/m**3)")

    def test_units_protected(self):
        def setunits(u, v):
            u.units = v

        def inplace(op, u, val):
            getattr(u, '__i%s__'%op)(val)

        self.assertRaises(AttributeError, setunits, pq.m, pq.ft)
        self.assertRaises(TypeError, inplace, 'add', pq.m, pq.m)
        self.assertRaises(TypeError, inplace, 'sub', pq.m, pq.m)
        self.assertRaises(TypeError, inplace, 'mul', pq.m, pq.m)
        self.assertRaises(TypeError, inplace, 'truediv', pq.m, pq.m)
        self.assertRaises(TypeError, inplace, 'pow', pq.m, 2)

    def test_units_copy(self):
        self.assertQuantityEqual(pq.m.copy(), pq.m)
        pc_per_cc = pq.CompoundUnit("pc/cm**3")
        self.assertQuantityEqual(pc_per_cc.copy(), pc_per_cc)



"""
"""

from ._utils import _cd
from ..unitquantity import UnitConstant


e = elementary_charge = UnitConstant(
    'elementary_charge',
    _cd('elementary charge'),
    symbol='e'
)
elementary_charge_over_h = UnitConstant(
    'elementary_charge_over_h',
    _cd('elementary charge over h'),
    symbol='e/h'
)
Faraday_constant = UnitConstant(
    'Faraday_constant',
    _cd('Faraday constant'),
    symbol='F'
)
#F_star = Faraday_constant_for_conventional_electric_current = UnitConstant(
#    _cd('Faraday constant for conventional electric current') what is a unit of C_90?
r_e = classical_electron_radius = UnitConstant(
    'classical_electron_radius',
    _cd('classical electron radius'),
    symbol='r_e',
    u_symbol='rₑ'
)
m_e = electron_mass = UnitConstant(
    'electron_mass',
    _cd('electron mass'),
    symbol='m_e',
    u_symbol='mₑ'
)
lambda_C = Compton_wavelength = UnitConstant(
    'Compton_wavelength',
    _cd('Compton wavelength'),
    symbol='lambda_C',
    u_symbol='λ_C'
)
Compton_wavelength_over_2_pi = UnitConstant(
    'Compton_wavelength_over_2_pi',
    _cd('Compton wavelength over 2 pi'),
    symbol='lambdabar_C',
    u_symbol='ƛ_C'
)
electron_charge_to_mass_quotient = UnitConstant(
    'electron_charge_to_mass_quotient',
    _cd('electron charge to mass quotient'),
    symbol='(-e/m_e)',
    u_symbol='(-e/mₑ)'
)
g_e = electron_g_factor = UnitConstant(
    'electron_g_factor',
    _cd('electron g factor'),
    symbol='g_e',
    u_symbol='gₑ'
)
gamma_e = electron_gyromagnetic_ratio = UnitConstant(
    'electron_gyromagnetic_ratio',
    _cd('electron gyromagnetic ratio'),
    symbol='gamma_e',
    u_symbol='γₑ'
)
electron_gyromagnetic_ratio_over_2_pi = UnitConstant(
    'electron_gyromagnetic_ratio_over_2_pi',
    _cd('electron gyromagnetic ratio over 2 pi'),
    symbol='gamma_e/(2*pi)',
    u_symbol='γₑ/(2·π)'
)
mu_e = electron_magnetic_moment = UnitConstant(
    'electron_magnetic_moment',
    _cd('electron magnetic moment'),
    symbol='mu_e',
    u_symbol='μₑ'
)
a_e = electron_magnetic_moment_anomaly = UnitConstant(
    'electron_magnetic_moment_anomaly',
    _cd('electron magnetic moment anomaly'),
    symbol='a_e',
    u_symbol='aₑ'
)
eV = electron_volt = UnitConstant(
    'electron_volt',
    _cd('electron volt'),
    symbol='eV'
)
sigma_e = Thomson_cross_section = UnitConstant(
    'Thomson_cross_section',
    _cd('Thomson cross section'),
    symbol='sigma_e',
    u_symbol='σₑ'
)

mu_B = Bohr_magneton = UnitConstant(
    'Bohr_magneton',
    _cd('Bohr magneton'),
    symbol='mu_B',
    u_symbol='μ_B'
)
Bohr_magneton_in_Hz_per_T = UnitConstant(
    'Bohr_magneton_in_Hz_per_T',
    _cd('Bohr magneton in Hz/T')
)
Bohr_magneton_in_inverse_meters_per_tesla = UnitConstant(
    'Bohr_magneton_in_inverse_meters_per_tesla',
    _cd('Bohr magneton in inverse meters per tesla')
)
Bohr_magneton_in_K_per_T = UnitConstant(
    'Bohr_magneton_in_K_per_T',
    _cd('Bohr magneton in K/T')
)

electron_mass_energy_equivalent = UnitConstant(
    'electron_mass_energy_equivalent',
    _cd('electron mass energy equivalent'),
    symbol='(m_e*c**2)',
    u_symbol='(mₑ·c²)'
)
electron_mass_energy_equivalent_in_MeV = UnitConstant(
    'electron_mass_energy_equivalent_in_MeV',
    _cd('electron mass energy equivalent in MeV')
)
electron_mass_in_u = UnitConstant(
    'electron_mass_in_u',
    _cd('electron mass in u')
)
electron_molar_mass = UnitConstant(
    'electron_molar_mass',
    _cd('electron molar mass'),
    symbol='M_e',
    u_symbol='Mₑ'
)

electron_deuteron_mass_ratio = UnitConstant(
    'electron_deuteron_mass_ratio',
    _cd('electron-deuteron mass ratio'),
    symbol='(m_e/m_d)',
    u_symbol='(mₑ/m_d)'
)
electron_muon_mass_ratio = UnitConstant(
    'electron_muon_mass_ratio',
    _cd('electron-muon mass ratio'),
    symbol='(m_e/m_mu)',
    u_symbol='(mₑ/m_μ)'
)
electron_neutron_mass_ratio = UnitConstant(
    'electron_neutron_mass_ratio',
    _cd('electron-neutron mass ratio'),
    symbol='(m_e/m_n)',
    u_symbol='(mₑ/m_n)'
)
electron_proton_mass_ratio = UnitConstant(
    'electron_proton_mass_ratio',
    _cd('electron-proton mass ratio'),
    symbol='(m_e/m_p)',
    u_symbol='(mₑ/m_p)'
)
electron_tau_mass_ratio = UnitConstant(
    'electron_tau_mass_ratio',
    _cd('electron-tau mass ratio'),
    symbol='(m_e/m_tau)',
    u_symbol='(mₑ/m_τ)'
)
electron_to_alpha_particle_mass_ratio = UnitConstant(
    'electron_to_alpha_particle_mass_ratio',
    _cd('electron to alpha particle mass ratio'),
    symbol='(m_e/m_alpha)',
    u_symbol='(mₑ/m_α)'
)

electron_deuteron_magnetic_moment_ratio = UnitConstant(
    'electron_deuteron_magnetic_moment_ratio',
    _cd('electron-deuteron magnetic moment ratio'),
    symbol='(mu_e/mu_d)',
    u_symbol='(μₑ/μ_d)'
)
electron_magnetic_moment_to_Bohr_magneton_ratio = UnitConstant(
    'electron_magnetic_moment_to_Bohr_magneton_ratio',
    _cd('electron magnetic moment to Bohr magneton ratio'),
    symbol='(mu_e/mu_B)',
    u_symbol='(μₑ/μ_B)'
)
electron_magnetic_moment_to_nuclear_magneton_ratio = UnitConstant(
    'electron_magnetic_moment_to_nuclear_magneton_ratio',
    _cd('electron magnetic moment to nuclear magneton ratio'),
    symbol='(mu_e/mu_N)',
    u_symbol='(μₑ/μ_N)'
)
electron_muon_magnetic_moment_ratio = UnitConstant(
    'electron_muon_magnetic_moment_ratio',
    _cd('electron-muon magnetic moment ratio'),
    symbol='(mu_e/mu_mu)',
    u_symbol='(μₑ/μ_μ)'
)
electron_neutron_magnetic_moment_ratio = UnitConstant(
    'electron_neutron_magnetic_moment_ratio',
    _cd('electron-neutron magnetic moment ratio'),
    symbol='(mu_e/mu_n)',
    u_symbol='(μₑ/μ_n)'
)
electron_proton_magnetic_moment_ratio = UnitConstant(
    'electron_proton_magnetic_moment_ratio',
    _cd('electron-proton magnetic moment ratio'),
    symbol='(mu_e/mu_p)',
    u_symbol='(μₑ/μ_p)'
)
electron_to_shielded_helion_magnetic_moment_ratio = UnitConstant(
    'electron_to_shielded_helion_magnetic_moment_ratio',
    _cd('electron to shielded helion magnetic moment ratio'),
    symbol='(mu_e/muprime_h)',
    u_symbol='(μₑ/μ′_h)'
)
electron_to_shielded_proton_magnetic_moment_ratio = UnitConstant(
    'electron_to_shielded_proton_magnetic_moment_ratio',
    _cd('electron to shielded proton magnetic moment ratio'),
    symbol='(mu_e/muprime_p)',
    u_symbol='(μₑ/μ′_p)'
)


electron_volt_atomic_mass_unit_relationship = UnitConstant(
    'electron_volt_atomic_mass_unit_relationship',
    _cd('electron volt-atomic mass unit relationship')
)
electron_volt_hartree_relationship = UnitConstant(
    'electron_volt_hartree_relationship',
    _cd('electron volt-hartree relationship')
)
electron_volt_hertz_relationship = UnitConstant(
    'electron_volt_hertz_relationship',
    _cd('electron volt-hertz relationship')
)
electron_volt_inverse_meter_relationship = UnitConstant(
    'electron_volt_inverse_meter_relationship',
    _cd('electron volt-inverse meter relationship')
)
electron_volt_joule_relationship = UnitConstant(
    'electron_volt_joule_relationship',
    _cd('electron volt-joule relationship')
)
electron_volt_kelvin_relationship = UnitConstant(
    'electron_volt_kelvin_relationship',
    _cd('electron volt-kelvin relationship')
)
electron_volt_kilogram_relationship = UnitConstant(
    'electron_volt_kilogram_relationship',
    _cd('electron volt-kilogram relationship')
)
hertz_electron_volt_relationship = UnitConstant(
    'hertz_electron_volt_relationship',
    _cd('hertz-electron volt relationship')
)
inverse_meter_electron_volt_relationship = UnitConstant(
    'inverse_meter_electron_volt_relationship',
    _cd('inverse meter-electron volt relationship')
)
joule_electron_volt_relationship = UnitConstant(
    'joule_electron_volt_relationship',
    _cd('joule-electron volt relationship')
)
kelvin_electron_volt_relationship = UnitConstant(
    'kelvin_electron_volt_relationship',
    _cd('kelvin-electron volt relationship')
)
kilogram_electron_volt_relationship = UnitConstant(
    'kilogram_electron_volt_relationship',
    _cd('kilogram-electron volt relationship')
)

del UnitConstant, _cd



"""
"""

from ._utils import _cd
from ..unitquantity import UnitConstant


natural_unit_of_action = UnitConstant(
    'natural_unit_of_action',
    _cd('natural unit of action'),
    symbol='hbar',
    u_symbol='ħ'
)
natural_unit_of_energy = UnitConstant(
    'natural_unit_of_energy',
    _cd('natural unit of energy'),
    symbol='(m_e*c**2)',
    u_symbol='(mₑ·c²)'
)
natural_unit_of_length = UnitConstant(
    'natural_unit_of_length',
    _cd('natural unit of length'),
    symbol='lambdabar_C',
    u_symbol='ƛ_C'
)
natural_unit_of_mass = UnitConstant(
    'natural_unit_of_mass',
    _cd('natural unit of mass'),
    symbol='m_e',
    u_symbol='mₑ'
)
natural_unit_of_momentum = UnitConstant(
    'natural_unit_of_momentum',
    _cd('natural unit of momentum'),
    symbol='(m_e*c)',
    u_symbol='(mₑ·c)'
)
natural_unit_of_time = UnitConstant(
    'natural_unit_of_time',
    _cd('natural unit of time'),
    symbol='(hbar/(m_e*c**2))',
    u_symbol='(ħ/(mₑ·c²))'
)
natural_unit_of_velocity = UnitConstant(
    'natural_unit_of_velocity',
    _cd('natural unit of velocity'),
    symbol='c'
)

natural_unit_of_action_in_eV_s = UnitConstant(
    'natural_unit_of_action_in_eV_s',
    _cd('natural unit of action in eV s')
)
natural_unit_of_energy_in_MeV = UnitConstant(
    'natural_unit_of_energy_in_MeV',
    _cd('natural unit of energy in MeV')
)
natural_unit_of_momentum_in_MeV_per_c = UnitConstant(
    'natural_unit_of_momentum_in_MeV_per_c',
    _cd('natural unit of momentum in MeV/c')
)

del UnitConstant, _cd



"""
"""

from ..unitquantity import UnitQuantity
from .length import cm, m, foot, inch
from .area import acre

l = L = liter = litre = UnitQuantity(
    'liter',
    1e-3*m**3,
    symbol='L',
    aliases=['l', 'liters', 'litre', 'litres']
)
mL = milliliter = millilitre = UnitQuantity(
    'milliliter',
    liter/1000,
    symbol='mL',
    aliases=['ml', 'milliliters', 'millilitre', 'millilitres']
)
kL = kiloliter = kilolitre = UnitQuantity(
    'kiloliter',
    liter*1000,
    symbol='kL',
    aliases=['kl', 'kiloliters', 'kilolitre', 'kilolitres']
)
ML = megaliter = megalitre = UnitQuantity(
    'megaliter',
    kiloliter*1000,
    symbol='ML',
    aliases=['Ml', 'megaliters', 'megalitre', 'megalitres']
)
GL = gigaliter = gigalitre = UnitQuantity(
    'gigaliter',
    megaliter*1000,
    symbol='GL',
    aliases=['Gl', 'gigaliters', 'gigalitre', 'gigalitres']
)
cc = cubic_centimeter = milliliter = UnitQuantity(
    'cubic_centimeter',
    cm**3,
    symbol='cc',
    aliases=['cubic_centimeters']
)

stere = UnitQuantity(
    'stere',
    m**3,
    aliases=['steres']
)
gross_register_ton = register_ton = UnitQuantity(
    'gross_register_ton',
    100*foot**3,
    symbol='GRT',
    aliases=['gross_register_tons', 'register_ton', 'register_tons']
)

acre_foot = UnitQuantity(
    'acre_foot',
    acre*foot,
    aliases=['acre_feet']
)
board_foot = UnitQuantity(
    'board_foot',
    foot**2*inch,
    symbol='FBM',
    aliases=['board_feet'])
bu = bushel = US_bushel = UnitQuantity(
    'US_bushel',
    2150.42*inch**3,
    symbol='bu'
)
US_dry_gallon = UnitQuantity(
    'US_dry_gallon',
    bushel/8,
)
gallon = liquid_gallon = US_liquid_gallon = UnitQuantity(
    'US_liquid_gallon',
    231*inch**3,
    aliases=[
        'gallon', 'gallons', 'liquid_gallon', 'liquid_gallons',
        'US_liquid_gallons'
    ],
)
dry_quart = US_dry_quart = UnitQuantity(
    'US_dry_quart',
    US_dry_gallon/4,
    aliases=['dry_quart', 'dry_quarts', 'US_dry_quarts']
)
dry_pint = US_dry_pint = UnitQuantity(
    'US_dry_pint',
    US_dry_quart/2,
    aliases=['dry_pint', 'dry_pints', 'US_dry_pints']
)
quart = liquid_quart = US_liquid_quart = UnitQuantity(
    'US_liquid_quart',
    US_liquid_gallon/4,
    symbol='quart',
    aliases=['quarts', 'liquid_quart', 'liquid_quarts', 'US_liquid_quarts']
)
pt = pint = liquid_pint = US_liquid_pint = UnitQuantity(
    'US_liquid_pint',
    US_liquid_quart/2,
    symbol='pt',
    aliases=[
        'pint', 'pints', 'liquid_pint', 'liquid_pints', 'US_liquid_pints'
    ],
)
cup = US_liquid_cup = UnitQuantity(
    'cup',
    US_liquid_pint/2,
    aliases=['cups', 'US_liquid_cup', 'US_liquid_cups']
)
gill = US_liquid_gill = UnitQuantity(
    'US_liquid_gill',
    US_liquid_cup/2,
    symbol='gill',
    aliases=['gills', 'US_liquid_gills']
)
floz = fluid_ounce = US_fluid_ounce = US_liquid_ounce = UnitQuantity(
    'US_fluid_ounce',
    US_liquid_gill/4,
    symbol='fl_oz',
    aliases=[
        'fluid_ounce', 'fluid_ounces', 'US_fluid_ounces', 'US_liquid_ounce',
        'US_liquid_ounces'
    ]
)

Imperial_bushel = UnitQuantity(
    'Imperial_bushel',
    36.36872*liter,
    doc='exact'
)
UK_liquid_gallon = Canadian_liquid_gallon = UnitQuantity(
    'UK_liquid_gallon',
    4.54609*liter,
    aliases=[
        'UK_liquid_gallons', 'Canadian_liquid_gallon',
        'Canadian_liquid_gallons'
    ],
    doc='exact'
)
UK_liquid_quart = UnitQuantity(
    'UK_liquid_quart',
    UK_liquid_gallon/4,
    aliases=['UK_liquid_quarts']
)
UK_liquid_pint = UnitQuantity(
    'UK_liquid_pint',
    UK_liquid_quart/2,
    aliases=['UK_liquid_pints']
)
UK_liquid_cup = UnitQuantity(
    'UK_liquid_cup',
    UK_liquid_pint/2,
    aliases=['UK_liquid_cups']
)
UK_liquid_gill = UnitQuantity(
    'UK_liquid_gill',
    UK_liquid_cup/2,
    aliases=['UK_liquid_gills']
)
UK_fluid_ounce = UK_liquid_ounce = UnitQuantity(
    'UK_fluid_ounce',
    UK_liquid_gill/5, # not a mistake
    aliases=['UK_fluid_ounces', 'UK_liquid_ounce', 'UK_liquid_ounces']
)

bbl = barrel = UnitQuantity(
    'barrel',
    42*US_liquid_gallon,
    symbol='bbl'
)
tbsp = Tbsp = Tblsp = tblsp = tbs = Tbl = tablespoon = UnitQuantity(
    'tablespoon',
    US_fluid_ounce/2,
    symbol='tbsp',
    aliases=['Tbsp', 'Tblsp', 'Tbl', 'tblsp', 'tbs', 'tablespoons']
)
tsp = teaspoon = UnitQuantity(
    'teaspoon',
    tablespoon/3,
    symbol='tsp',
    aliases=['teaspoons']
)

pk = peck = UnitQuantity(
    'peck',
    bushel/4,
    symbol='pk',
    aliases=['pecks']
)

fldr = fluid_dram = fluidram = UnitQuantity(
    'fluid_dram',
    floz/8,
    symbol='fldr',
    aliases=['fluid_drams', 'fluidram', 'fluidrams']
)

firkin = UnitQuantity(
    'firkin',
    barrel/4
)

del UnitQuantity, cm, m



"""
"""

from ._utils import _cd
from ..unitquantity import UnitConstant


m_n = neutron_mass = UnitConstant(
    'neutron_mass',
    _cd('neutron mass'),
    symbol='m_n'
)
lambda_C_n = neutron_Compton_wavelength = UnitConstant(
    'neutron_Compton_wavelength',
    _cd('neutron Compton wavelength'),
    symbol='lambda_C_n',
    u_symbol='λ_C_n'
)
neutron_Compton_wavelength_over_2_pi = UnitConstant(
    'neutron_Compton_wavelength_over_2_pi',
    _cd('neutron Compton wavelength over 2 pi'),
    symbol='lambdabar_C_n',
    u_symbol='ƛ_C_n'
)
g_n = neutron_g_factor = UnitConstant(
    'neutron_g_factor',
    _cd('neutron g factor'),
    symbol='g_n'
)
gamma_n = neutron_gyromagnetic_ratio = UnitConstant(
    'neutron_gyromagnetic_ratio',
    _cd('neutron gyromagnetic ratio'),
    symbol='gamma_n',
    u_symbol='γ_n'
)
neutron_gyromagnetic_ratio_over_2_pi = UnitConstant(
    'neutron_gyromagnetic_ratio_over_2_pi',
    _cd('neutron gyromagnetic ratio over 2 pi'),
    symbol='(gamma_n/(2*pi))',
    u_symbol='(γ_n/(2·π))'
)
mu_n = neutron_magnetic_moment = UnitConstant(
    'neutron_magnetic_moment',
    _cd('neutron magnetic moment'),
    symbol='mu_n',
    u_symbol='μ_n'
)
neutron_mass_energy_equivalent = UnitConstant(
    'neutron_mass_energy_equivalent',
    _cd('neutron mass energy equivalent'),
    symbol='(m_n*c**2)',
    u_symbol='(m_n·c²)'
)
neutron_mass_energy_equivalent_in_MeV = UnitConstant(
    'neutron_mass_energy_equivalent_in_MeV',
    _cd('neutron mass energy equivalent in MeV')
)
neutron_mass_in_u = UnitConstant(
    'neutron_mass_in_u',
    _cd('neutron mass in u')
)
neutron_molar_mass = UnitConstant(
    'neutron_molar_mass',
    _cd('neutron molar mass'),
    symbol='M_n'
)

neutron_electron_mass_ratio = UnitConstant(
    'neutron_electron_mass_ratio',
    _cd('neutron-electron mass ratio'),
    symbol='(m_n/m_e)',
    u_symbol='(m_n/mₑ)'
)
neutron_muon_mass_ratio = UnitConstant(
    'neutron_muon_mass_ratio',
    _cd('neutron-muon mass ratio'),
    symbol='(m_n/m_mu)',
    u_symbol='(m_n/m_μ)'
)
neutron_proton_mass_ratio = UnitConstant(
    'neutron_proton_mass_ratio',
    _cd('neutron-proton mass ratio'),
    symbol='(m_n/m_p)',
    u_symbol='(m_n/m_p)'
)
neutron_tau_mass_ratio = UnitConstant(
    'neutron_tau_mass_ratio',
    _cd('neutron-tau mass ratio'),
    symbol='(m_n/m_tau)',
    u_symbol='(m_n/m_τ)'
)


neutron_electron_magnetic_moment_ratio = UnitConstant(
    'neutron_electron_magnetic_moment_ratio',
    _cd('neutron-electron magnetic moment ratio'),
    symbol='(mu_n/mu_e)',
    u_symbol='(μ_n/μₑ)'
)
neutron_magnetic_moment_to_Bohr_magneton_ratio = UnitConstant(
    'neutron_magnetic_moment_to_Bohr_magneton_ratio',
    _cd('neutron magnetic moment to Bohr magneton ratio'),
    symbol='(mu_n/mu_B)',
    u_symbol='(μ_n/μ_B)'
)
neutron_magnetic_moment_to_nuclear_magneton_ratio = UnitConstant(
    'neutron_magnetic_moment_to_nuclear_magneton_ratio',
    _cd('neutron magnetic moment to nuclear magneton ratio'),
    symbol='(mu_n/mu_N)',
    u_symbol='(μ_n/μ_N)'
)
neutron_proton_magnetic_moment_ratio = UnitConstant(
    'neutron_proton_magnetic_moment_ratio',
    _cd('neutron-proton magnetic moment ratio'),
    symbol='(mu_n/mu_p)',
    u_symbol='(μ_n/μ_p)'
)
neutron_to_shielded_proton_magnetic_moment_ratio = UnitConstant(
    'neutron_to_shielded_proton_magnetic_moment_ratio',
    _cd('neutron to shielded proton magnetic moment ratio'),
    symbol='(mu_n/muprime_p)',
    u_symbol='(μ_n/μ′_p)'
)

del UnitConstant, _cd



#SI prefixes
yotta = 1e24
zetta = 1e21
exa = 1e18
peta = 1e15
tera = 1e12
giga = 1e9
mega = 1e6
kilo = 1e3
hecto = 1e2
deka = 1e1
deci = 1e-1
centi = 1e-2
milli = 1e-3
micro = 1e-6
nano = 1e-9
pico = 1e-12
femto = 1e-15
atto = 1e-18
zepto = 1e-21

#binary prefixes
kibi = 2**10
mebi = 2**20
gibi = 2**30
tebi = 2**40
pebi = 2**50
exbi = 2**60
zebi = 2**70
yobi = 2**80



"""
"""

from ..unitquantity import UnitQuantity
from .temperature import K, degF
from .length import m, ft
from .power import W
from .energy import BTU
from .time import h


RSI = UnitQuantity(
    'RSI',
    K*m**2/W,
    doc='R-value in SI'
)

clo = clos = UnitQuantity(
    'clo',
    0.155*RSI,
    aliases=['clos']
)

R_value = UnitQuantity(
    'R_value',
    ft**2*degF*h/BTU,
    doc='American customary units'
)


del UnitQuantity, K, degF, m, ft, W, BTU, h



"""
"""

from ._utils import _cd
from ..unitquantity import UnitConstant


m_mu = muon_mass = UnitConstant(
    'muon_mass',
    _cd('muon mass'),
    symbol='m_mu',
    u_symbol='m_μ'
)
lambda_C_mu = muon_Compton_wavelength = UnitConstant(
    'muon_Compton_wavelength',
    _cd('muon Compton wavelength'),
    symbol='lambdabar_C_mu',
    u_symbol='λ_C_μ'
)
muon_Compton_wavelength_over_2_pi = UnitConstant(
    'muon_Compton_wavelength_over_2_pi',
    _cd('muon Compton wavelength over 2 pi'),
    symbol='lambdabar_Cmu',
    u_symbol='ƛ_C_μ'
)
g_mu = muon_g_factor = UnitConstant(
    'muon_g_factor',
    _cd('muon g factor'),
    symbol='g_mu',
    u_symbol='g_μ'
)
mu_mu = muon_magnetic_moment = UnitConstant(
    'muon_magnetic_moment',
    _cd('muon magnetic moment'),
    symbol='mu_mu',
    u_symbol='μ_μ'
)
a_mu = muon_magnetic_moment_anomaly = UnitConstant(
    'muon_magnetic_moment_anomaly',
    _cd('muon magnetic moment anomaly'),
    symbol='a_mu',
    u_symbol='a_μ'
)
muon_mass_energy_equivalent = UnitConstant(
    'muon_mass_energy_equivalent',
    _cd('muon mass energy equivalent'),
    symbol='(m_mu*c**2)',
    u_symbol='(m_μ·c²)'
)
muon_mass_energy_equivalent_in_MeV = UnitConstant(
    'muon_mass_energy_equivalent_in_MeV',
    _cd('muon mass energy equivalent in MeV')
)
muon_mass_in_u = UnitConstant(
    'muon_mass_in_u',
    _cd('muon mass in u')
)
muon_molar_mass = UnitConstant(
    'muon_molar_mass',
    _cd('muon molar mass'),
    symbol='M_mu',
    u_symbol='M_μ'
)

muon_electron_mass_ratio = UnitConstant(
    'muon_electron_mass_ratio',
    _cd('muon-electron mass ratio'),
    symbol='(m_mu/m_e)',
    u_symbol='(m_μ/mₑ)'
)
muon_neutron_mass_ratio = UnitConstant(
    'muon_neutron_mass_ratio',
    _cd('muon-neutron mass ratio'),
    symbol='(m_mu/m_n)',
    u_symbol='(m_μ/m_n)'
)
muon_proton_mass_ratio = UnitConstant(
    'muon_proton_mass_ratio',
    _cd('muon-proton mass ratio'),
    symbol='(m_mu/m_p)',
    u_symbol='(m_μ/m_p)'
)
muon_tau_mass_ratio = UnitConstant(
    'muon_tau_mass_ratio',
    _cd('muon-tau mass ratio'),
    symbol='(m_mu/m_tau)',
    u_symbol='(m_μ/m_τ)'
)

muon_magnetic_moment_to_Bohr_magneton_ratio = UnitConstant(
    'muon_magnetic_moment_to_Bohr_magneton_ratio',
    _cd('muon magnetic moment to Bohr magneton ratio'),
    symbol='(mu_mu/mu_B)',
    u_symbol='(μ_μ/μ_B)'
)
muon_magnetic_moment_to_nuclear_magneton_ratio = UnitConstant(
    'muon_magnetic_moment_to_nuclear_magneton_ratio',
    _cd('muon magnetic moment to nuclear magneton ratio'),
    symbol='(mu_mu/mu_N)',
    u_symbol='(μ_μ/μ_N)'
)
muon_proton_magnetic_moment_ratio = UnitConstant(
    'muon_proton_magnetic_moment_ratio',
    _cd('muon-proton magnetic moment ratio'),
    symbol='(mu_mu/mu_p)',
    u_symbol='(μ_μ/μ_p)'
)

del UnitConstant, _cd



"""
"""

from ..unitquantity import UnitQuantity
from .time import s
from .length import m

g_0 = g_n = gravity = standard_gravity = gee = force = free_fall = \
    standard_free_fall = gp = dynamic = geopotential = UnitQuantity(
    'standard_gravity',
    9.806650*m/s**2,
    symbol='g_0',
    u_symbol='g₀',
    doc='exact'
)

del m, s, UnitQuantity



"""
"""

from ._utils import _cd
from ..unitquantity import UnitConstant



m_h = helion_mass = UnitConstant(
    'helion_mass',
    _cd('helion mass'),
    symbol='m_h'
)
gamma_prime_h = shielded_helion_gyromagnetic_ratio = UnitConstant(
    'shielded_helion_gyromagnetic_ratio',
    _cd('shielded helion gyromagnetic ratio'),
    symbol='gammaprime_h',
    u_symbol='γ′_h'
)
shielded_helion_gyromagnetic_ratio_over_2_pi = UnitConstant(
    'shielded_helion_gyromagnetic_ratio_over_2_pi',
    _cd('shielded helion gyromagnetic ratio over 2 pi'),
    symbol='(gammaprime_h/(2*pi))',
    u_symbol='(γ′_h/(2·π))'
)
mu_prime_h = shielded_helion_magnetic_moment = UnitConstant(
    'shielded_helion_magnetic_moment',
    _cd('shielded helion magnetic moment'),
    symbol='muprime_h',
    u_symbol='μ′_h'
)
helion_mass_energy_equivalent = UnitConstant(
    'helion_mass_energy_equivalent',
    _cd('helion mass energy equivalent'),
    symbol='(m_h*c**2)',
    u_symbol='(m_h·c²)'
)
helion_mass_energy_equivalent_in_MeV = UnitConstant(
    'helion_mass_energy_equivalent_in_MeV',
    _cd('helion mass energy equivalent in MeV')
)
helion_mass_in_u = UnitConstant(
    'helion_mass_in_u',
    _cd('helion mass in u')
)
helion_molar_mass = UnitConstant(
    'helion_molar_mass',
    _cd('helion molar mass'),
    symbol='M_h'
)
helion_electron_mass_ratio = UnitConstant(
    'helion_electron_mass_ratio',
    _cd('helion-electron mass ratio'),
    symbol='(m_h/m_e)',
    u_symbol='(m_h/mₑ)'
)
helion_proton_mass_ratio = UnitConstant(
    'helion_proton_mass_ratio',
    _cd('helion-proton mass ratio'),
    symbol='(m_h/m_p)'
)

shielded_helion_to_proton_magnetic_moment_ratio = UnitConstant(
    'shielded_helion_to_proton_magnetic_moment_ratio',
    _cd('shielded helion to proton magnetic moment ratio'),
    symbol='(muprime_h/mu_p)',
    u_symbol='(μ′_h/μ_p)'
)
shielded_helion_magnetic_moment_to_Bohr_magneton_ratio = UnitConstant(
    'shielded_helion_magnetic_moment_to_Bohr_magneton_ratio',
    _cd('shielded helion magnetic moment to Bohr magneton ratio'),
    symbol='(muprime_h/mu_B)',
    u_symbol='(μ′_h/μ_B)'
)
shielded_helion_magnetic_moment_to_nuclear_magneton_ratio = UnitConstant(
    'shielded_helion_magnetic_moment_to_nuclear_magneton_ratio',
    _cd('shielded helion magnetic moment to nuclear magneton ratio'),
    symbol='(muprime_h/mu_N)',
    u_symbol='(μ′_h/μ_N)'
)
shielded_helion_to_shielded_proton_magnetic_moment_ratio = UnitConstant(
    'shielded_helion_to_shielded_proton_magnetic_moment_ratio',
    _cd('shielded helion to shielded proton magnetic moment ratio'),
    symbol='(muprime_h/muprime_p)',
    u_symbol='(μ′_h/μ′_p)'
)

del UnitConstant, _cd



from ._codata import physical_constants
from quantities.quantity import Quantity
from quantities.uncertainquantity import UncertainQuantity


def _cd(name):
    entry = physical_constants[name]
    if False: #entry['precision']:
        return UncertainQuantity(
            entry['value'], entry['units'], entry['precision']
        )
    else:
        return Quantity(entry['value'], entry['units'])



import operator as op

from .. import units as pq
from .common import TestCase


class TestComparison(TestCase):

    def test_scalar_equality(self):
        self.assertEqual(pq.J == pq.J, [True])
        self.assertEqual(1*pq.J == pq.J, [True])
        self.assertEqual(str(1*pq.J) == '1.0 J', True)
        self.assertEqual(pq.J == pq.kg*pq.m**2/pq.s**2, [True])

        self.assertEqual(pq.J == pq.erg, [False])
        self.assertEqual(2*pq.J == pq.J, [False])
        self.assertEqual(pq.J == 2*pq.kg*pq.m**2/pq.s**2, [False])

        self.assertEqual(pq.J == pq.kg, [False])

    def test_scalar_inequality(self):
        self.assertEqual(pq.J != pq.erg, [True])
        self.assertEqual(2*pq.J != pq.J, [True])
        self.assertEqual(str(2*pq.J) != str(pq.J), True)
        self.assertEqual(pq.J != 2*pq.kg*pq.m**2/pq.s**2, [True])

        self.assertEqual(pq.J != pq.J, [False])
        self.assertEqual(1*pq.J != pq.J, [False])
        self.assertEqual(pq.J != 1*pq.kg*pq.m**2/pq.s**2, [False])

    def test_scalar_comparison(self):
        self.assertEqual(2*pq.J > pq.J, [True])
        self.assertEqual(2*pq.J > 1*pq.J, [True])
        self.assertEqual(1*pq.J >= pq.J, [True])
        self.assertEqual(1*pq.J >= 1*pq.J, [True])
        self.assertEqual(2*pq.J >= pq.J, [True])
        self.assertEqual(2*pq.J >= 1*pq.J, [True])

        self.assertEqual(0.5*pq.J < pq.J, [True])
        self.assertEqual(0.5*pq.J < 1*pq.J, [True])
        self.assertEqual(0.5*pq.J <= pq.J, [True])
        self.assertEqual(0.5*pq.J <= 1*pq.J, [True])
        self.assertEqual(1.0*pq.J <= pq.J, [True])
        self.assertEqual(1.0*pq.J <= 1*pq.J, [True])

        self.assertEqual(2*pq.J < pq.J, [False])
        self.assertEqual(2*pq.J < 1*pq.J, [False])
        self.assertEqual(2*pq.J <= pq.J, [False])
        self.assertEqual(2*pq.J <= 1*pq.J, [False])

        self.assertEqual(0.5*pq.J > pq.J, [False])
        self.assertEqual(0.5*pq.J > 1*pq.J, [False])
        self.assertEqual(0.5*pq.J >= pq.J, [False])
        self.assertEqual(0.5*pq.J >= 1*pq.J, [False])

    def test_array_equality(self):
        self.assertQuantityEqual(
            [1, 2, 3, 4]*pq.J == [1, 22, 3, 44]*pq.J,
            [1, 0, 1, 0]
        )
        self.assertQuantityEqual(
            [1, 2, 3, 4]*pq.J == [1, 22, 3, 44]*pq.kg,
            [0, 0, 0, 0]
        )
        self.assertQuantityEqual(
            [1, 2, 3, 4]*pq.J == [1, 22, 3, 44],
            [1, 0, 1, 0]
        )

    def test_array_inequality(self):
        self.assertQuantityEqual(
            [1, 2, 3, 4]*pq.J != [1, 22, 3, 44]*pq.J,
            [0, 1, 0, 1]
        )
        self.assertQuantityEqual(
            [1, 2, 3, 4]*pq.J != [1, 22, 3, 44]*pq.kg,
            [1, 1, 1, 1]
        )
        self.assertQuantityEqual(
            [1, 2, 3, 4]*pq.J != [1, 22, 3, 44],
            [0, 1, 0, 1]
        )

    def test_quantity_less_than(self):
        self.assertQuantityEqual(
            [1, 2, 33]*pq.J < [1, 22, 3]*pq.J,
            [0, 1, 0]
        )
        self.assertQuantityEqual(
            [50, 100, 150]*pq.cm < [1, 1, 1]*pq.m,
            [1, 0, 0]
        )
        self.assertQuantityEqual(
            [1, 2, 33]*pq.J < [1, 22, 3],
            [0, 1, 0]
        )
        self.assertRaises(
            ValueError,
            op.lt,
            [1, 2, 33]*pq.J,
            [1, 22, 3]*pq.kg,
        )

    def test_quantity_less_than_or_equal(self):
        self.assertQuantityEqual(
            [1, 2, 33]*pq.J <= [1, 22, 3]*pq.J,
            [1, 1, 0]
        )
        self.assertQuantityEqual(
            [50, 100, 150]*pq.cm <= [1, 1, 1]*pq.m,
            [1, 1, 0]
        )
        self.assertQuantityEqual(
            [1, 2, 33]*pq.J <= [1, 22, 3],
            [1, 1, 0]
        )
        self.assertRaises(
            ValueError,
            op.le,
            [1, 2, 33]*pq.J,
            [1, 22, 3]*pq.kg,
        )

    def test_quantity_greater_than_or_equal(self):
        self.assertQuantityEqual(
            [1, 2, 33]*pq.J >= [1, 22, 3]*pq.J,
            [1, 0, 1]
        )
        self.assertQuantityEqual(
            [50, 100, 150]*pq.cm >= [1, 1, 1]*pq.m,
            [0, 1, 1]
        )
        self.assertQuantityEqual(
            [1, 2, 33]*pq.J >= [1, 22, 3],
            [1, 0, 1]
        )
        self.assertRaises(
            ValueError,
            op.ge,
            [1, 2, 33]*pq.J,
            [1, 22, 3]*pq.kg,
        )

    def test_quantity_greater_than(self):
        self.assertQuantityEqual(
            [1, 2, 33]*pq.J > [1, 22, 3]*pq.J,
            [0, 0, 1]
        )
        self.assertQuantityEqual(
            [50, 100, 150]*pq.cm > [1, 1, 1]*pq.m,
            [0, 0, 1]
        )
        self.assertQuantityEqual(
            [1, 2, 33]*pq.J > [1, 22, 3],
            [0, 0, 1]
        )
        self.assertRaises(
            ValueError,
            op.gt,
            [1, 2, 33]*pq.J,
            [1, 22, 3]*pq.kg,
        )



import operator as op
from functools import partial
import sys

import numpy as np
from .. import units as pq
from ..quantity import Quantity
from .common import TestCase


if sys.version.startswith('3'):
    long = int


def rand(dtype, *args):
    if np.dtype(dtype).kind == 'c':
        return dtype(
            10*np.random.rand(*args)+10j*np.random.rand(*args),
        )
    else:
        return dtype(10*np.random.rand(*args))

def check(f, *args, **kwargs):
    new = partial(f, *args)
    new.__name__ = f.__name__
    new.__module__ = f.__module__

    try:
        new = kwargs['fails_if'](new)
    except KeyError:
        pass
    desc = [f.__name__]
    for arg in args:
        try:
            desc.append(arg[0].dtype.name)
        except AttributeError:
            desc.append(arg[0].__class__.__name__)
        except (IndexError, TypeError):
            try:
                desc.append(arg.dtype.name)
            except AttributeError:
                pass
        c = arg.__class__.__name__
        if c != desc[-1]:
            desc.append(c)

    new.description = '_'.join(desc)
    return (new, )


class iter_dtypes:

    def __init__(self):
        self._i = 1
        self._typeDict = np.sctypeDict.copy()
        self._typeDict[17] = int
        self._typeDict[18] = long
        self._typeDict[19] = float
        self._typeDict[20] = complex

    def __iter__(self):
        return self

    def __next__(self):
        if self._i > 20:
            raise StopIteration

        i = self._i
        self._i += 1
        return self._typeDict[i]

    def next(self):
        return self.__next__()

def get_dtypes():
    return list(iter_dtypes())


class iter_types:

    def __init__(self, dtype):
        self._index = -1
        self._dtype = dtype

    def __iter__(self):
        return self

    def __next__(self):
        self._index += 1
        if self._index > 2:
            raise StopIteration
        if self._index > 0 and self._dtype in (int, long, float, complex):
            raise StopIteration
        if self._index == 0:
            return rand(self._dtype)
        if self._index == 1:
            return rand(self._dtype, 5).tolist()
        if self._index == 2:
            return rand(self._dtype, 5)

    def next(self):
        return self.__next__()


class TestDTypes(TestCase):

    def check_mul(self, m1, m2):
        self.assertQuantityEqual(pq.m*m2, Quantity(m2, 'm'))

        q1 = Quantity(m1, 'm')
        q2 = Quantity(m2, 's')
        a1 = np.asarray(m1)
        a2 = np.asarray(m2)
        self.assertQuantityEqual(q1*m2, Quantity(a1*a2, 'm'))
        self.assertQuantityEqual(q1*q2, Quantity(a1*a2, 'm*s'))

    def check_rmul(self, m1, m2):
        self.assertQuantityEqual(m1*pq.m, Quantity(m1, 'm'))

        q2 = Quantity(m2, 's')
        a1 = np.asarray(m1)
        a2 = np.asarray(m2)
        self.assertQuantityEqual(m1*q2, Quantity(a1*a2, 's'))

    def test_mul(self):
        dtypes = get_dtypes()
        for i in get_dtypes()[::3]:
            for j in get_dtypes()[::3]:
                for x in iter_types(i):
                    for y in iter_types(j):
                        self.check_mul(x, y)
                        self.check_rmul(x, y)
            dtypes.pop(0)

    def test_truediv(self):
        q = Quantity([44, 40, 36, 32], units=pq.ms)
        self.assertQuantityEqual(
            q/(4 * pq.ms),
            Quantity([11, 10, 9, 8], units=pq.dimensionless)
        )

    def test_floordiv(self):
        q = Quantity([45, 43, 39, 32], units=pq.ms)
        self.assertQuantityEqual(
            q//(4 * pq.ms),
            Quantity([11, 10, 9, 8], units=pq.dimensionless)
        )

    def test_mixed_addition(self):
        self.assertQuantityEqual(1*pq.ft + 1*pq.m, 4.280839895 * pq.ft)
        self.assertQuantityEqual(1*pq.ft + pq.m, 4.280839895 * pq.ft)
        self.assertQuantityEqual(pq.ft + 1*pq.m, 4.280839895 * pq.ft)
        self.assertQuantityEqual(pq.ft + pq.m, 4.280839895 * pq.ft)
        self.assertQuantityEqual(op.iadd(1*pq.ft, 1*pq.m), 4.280839895 * pq.ft)
        self.assertRaises(ValueError, lambda: 10*pq.J + 3*pq.m)
        self.assertRaises(ValueError, lambda: op.iadd(10*pq.J, 3*pq.m))

    def test_mod(self):
        self.assertQuantityEqual(10*pq.m % (3*pq.m), 1*pq.m)
        self.assertQuantityEqual(
            10*pq.m % (3*pq.m).rescale('ft'),
            10*pq.m % (3*pq.m)
        )
        self.assertRaises(ValueError, lambda: 10*pq.J % (3*pq.m))

    def test_imod(self):
        x = 10*pq.m
        x %= 3*pq.m
        self.assertQuantityEqual(x, 1*pq.m)

        x = 10*pq.m
        x %= (3*pq.m).rescale('ft')
        self.assertQuantityEqual(x, 10*pq.m % (3*pq.m))

        self.assertRaises(ValueError, lambda: op.imod(10*pq.J, 3*pq.m))

    def test_fmod(self):
        self.assertQuantityEqual(np.fmod(10*pq.m, (3*pq.m)), 1*pq.m)
        self.assertRaises(ValueError, np.fmod, 10*pq.J, 3*pq.m)

    def test_remainder(self):
        self.assertQuantityEqual(np.remainder(10*pq.m, (3*pq.m)), 1*pq.m)
        self.assertRaises(ValueError, np.remainder, 10*pq.J, 3*pq.m)

    def test_negative(self):
        self.assertQuantityEqual(
            -pq.m,
            Quantity(-1, 'm')
        )
        self.assertQuantityEqual(
            -Quantity(5, 'm'),
            Quantity(-5, 'm')
        )
        self.assertQuantityEqual(
            -Quantity(-5.0, 'm'),
            Quantity(5.0, 'm')
        )

    def test_addition(self):
        self.assertQuantityEqual(
            pq.eV + pq.eV,
            2*pq.eV
        )
        self.assertQuantityEqual(
            pq.eV + -pq.eV,
            0*pq.eV
        )
        self.assertQuantityEqual(
            pq.eV + 5*pq.eV,
            6*pq.eV
        )
        self.assertQuantityEqual(
            5*pq.eV + pq.eV,
            6*pq.eV
        )
        self.assertQuantityEqual(
            5*pq.eV + 6*pq.eV,
            11*pq.eV
        )
        self.assertQuantityEqual(
            pq.rem + [1, 2, 3]*pq.rem,
            [2, 3, 4]*pq.rem
        )
        self.assertQuantityEqual(
            [1, 2, 3]*pq.rem + pq.rem,
            [2, 3, 4]*pq.rem
        )
        self.assertQuantityEqual(
            5*pq.rem + [1, 2, 3]*pq.rem,
            [6, 7, 8]*pq.rem
        )
        self.assertQuantityEqual(
            [1, 2, 3]*pq.rem + 5*pq.rem,
            [6, 7, 8]*pq.rem
        )
        self.assertQuantityEqual(
            [1, 2, 3]*pq.hp + [1, 2, 3]*pq.hp,
            [2, 4, 6]*pq.hp
        )

        self.assertRaises(ValueError, op.add, pq.kPa, pq.lb)
        self.assertRaises(ValueError, op.add, pq.kPa, 10)
        self.assertRaises(ValueError, op.add, 1*pq.kPa, 5*pq.lb)
        self.assertRaises(ValueError, op.add, 1*pq.kPa, pq.lb)
        self.assertRaises(ValueError, op.add, 1*pq.kPa, 5)
        self.assertRaises(ValueError, op.add, [1, 2, 3]*pq.kPa, [1, 2, 3]*pq.lb)
        self.assertRaises(ValueError, op.add, [1, 2, 3]*pq.kPa, 5*pq.lb)
        self.assertRaises(ValueError, op.add, [1, 2, 3]*pq.kPa, pq.lb)
        self.assertRaises(ValueError, op.add, [1, 2, 3]*pq.kPa, 5)

    def test_in_place_addition(self):
        x = 1*pq.m
        x += pq.m
        self.assertQuantityEqual(x, pq.m+pq.m)

        x = 1*pq.m
        x += -pq.m
        self.assertQuantityEqual(x, 0*pq.m)

        x = [1, 2, 3, 4]*pq.m
        x += pq.m
        self.assertQuantityEqual(x, [2, 3, 4, 5]*pq.m)

        x = [1, 2, 3, 4]*pq.m
        x += x
        self.assertQuantityEqual(x, [2, 4, 6, 8]*pq.m)

        x = [1, 2, 3, 4]*pq.m
        x[:2] += pq.m
        self.assertQuantityEqual(x, [2, 3, 3, 4]*pq.m)

        x = [1, 2, 3, 4]*pq.m
        x[:2] += -pq.m
        self.assertQuantityEqual(x, [0, 1, 3, 4]*pq.m)

        x = [1, 2, 3, 4]*pq.m
        x[:2] += [1, 2]*pq.m
        self.assertQuantityEqual(x, [2, 4, 3, 4]*pq.m)

        x = [1, 2, 3, 4]*pq.m
        x[::2] += [1, 2]*pq.m
        self.assertQuantityEqual(x, [2, 2, 5, 4]*pq.m)

        self.assertRaises(ValueError, op.iadd, 1*pq.m, 1)
        self.assertRaises(ValueError, op.iadd, 1*pq.m, pq.J)
        self.assertRaises(ValueError, op.iadd, 1*pq.m, 5*pq.J)
        self.assertRaises(ValueError, op.iadd, [1, 2, 3]*pq.m, 1)
        self.assertRaises(ValueError, op.iadd, [1, 2, 3]*pq.m, pq.J)
        self.assertRaises(ValueError, op.iadd, [1, 2, 3]*pq.m, 5*pq.J)

    def test_subtraction(self):
        self.assertQuantityEqual(
            pq.eV - pq.eV,
            0*pq.eV
        )
        self.assertQuantityEqual(
            5*pq.eV - pq.eV,
            4*pq.eV
        )
        self.assertQuantityEqual(
            pq.eV - 4*pq.eV,
            -3*pq.eV
        )
        self.assertQuantityEqual(
            pq.rem - [1, 2, 3]*pq.rem,
            [0, -1, -2]*pq.rem
        )
        self.assertQuantityEqual(
            [1, 2, 3]*pq.rem - pq.rem,
            [0, 1, 2]*pq.rem
        )
        self.assertQuantityEqual(
            5*pq.rem - [1, 2, 3]*pq.rem,
            [4, 3, 2]*pq.rem
        )
        self.assertQuantityEqual(
            [1, 2, 3]*pq.rem - 5*pq.rem,
            [-4, -3, -2]*pq.rem
        )
        self.assertQuantityEqual(
            [3, 3, 3]*pq.hp - [1, 2, 3]*pq.hp,
            [2, 1, 0]*pq.hp
        )

        self.assertRaises(ValueError, op.sub, pq.kPa, pq.lb)
        self.assertRaises(ValueError, op.sub, pq.kPa, 10)
        self.assertRaises(ValueError, op.sub, 1*pq.kPa, 5*pq.lb)
        self.assertRaises(ValueError, op.sub, 1*pq.kPa, pq.lb)
        self.assertRaises(ValueError, op.sub, 1*pq.kPa, 5)
        self.assertRaises(ValueError, op.sub, [1, 2, 3]*pq.kPa, [1, 2, 3]*pq.lb)
        self.assertRaises(ValueError, op.sub, [1, 2, 3]*pq.kPa, 5*pq.lb)
        self.assertRaises(ValueError, op.sub, [1, 2, 3]*pq.kPa, pq.lb)
        self.assertRaises(ValueError, op.sub, [1, 2, 3]*pq.kPa, 5)

    def test_in_place_subtraction(self):
        x = 1*pq.m
        x -= pq.m
        self.assertQuantityEqual(x, 0*pq.m)

        x = 1*pq.m
        x -= -pq.m
        self.assertQuantityEqual(x, 2*pq.m)

        x = [1, 2, 3, 4]*pq.m
        x -= pq.m
        self.assertQuantityEqual(x, [0, 1, 2, 3]*pq.m)

        x = [1, 2, 3, 4]*pq.m
        x -= [1, 1, 1, 1]*pq.m
        self.assertQuantityEqual(x, [0, 1, 2, 3]*pq.m)

        x = [1, 2, 3, 4]*pq.m
        x[:2] -= pq.m
        self.assertQuantityEqual(x, [0, 1, 3, 4]*pq.m)

        x = [1, 2, 3, 4]*pq.m
        x[:2] -= -pq.m
        self.assertQuantityEqual(x, [2, 3, 3, 4]*pq.m)

        x = [1, 2, 3, 4]*pq.m
        x[:2] -= [1, 2]*pq.m
        self.assertQuantityEqual(x, [0, 0, 3, 4]*pq.m)

        x = [1, 2, 3, 4]*pq.m
        x[::2] -= [1, 2]*pq.m
        self.assertQuantityEqual(x, [0, 2, 1, 4]*pq.m)

        self.assertRaises(ValueError, op.isub, 1*pq.m, 1)
        self.assertRaises(ValueError, op.isub, 1*pq.m, pq.J)
        self.assertRaises(ValueError, op.isub, 1*pq.m, 5*pq.J)
        self.assertRaises(ValueError, op.isub, [1, 2, 3]*pq.m, 1)
        self.assertRaises(ValueError, op.isub, [1, 2, 3]*pq.m, pq.J)
        self.assertRaises(ValueError, op.isub, [1, 2, 3]*pq.m, 5*pq.J)

    def test_division(self):
        molar = pq.UnitQuantity('M',  1000 * pq.mole/pq.m**3, u_symbol='M')
        for subtr in [1, 1.0]:
            q = 1*molar/(1000*pq.mole/pq.m**3)
            self.assertQuantityEqual((q - subtr).simplified, 0)

    def test_powering(self):
        # test raising a quantity to a power
        self.assertQuantityEqual((5.5 * pq.cm)**5, (5.5**5) * (pq.cm**5))
        self.assertQuantityEqual((5.5 * pq.cm)**0, (5.5**0) * pq.dimensionless)

        # must also work with compound units
        self.assertQuantityEqual((5.5 * pq.J)**5, (5.5**5) * (pq.J**5))

        # does powering work with arrays?
        self.assertQuantityEqual(
            (np.array([1, 2, 3, 4, 5]) * pq.kg)**3,
            np.array([1, 8, 27, 64, 125]) * pq.kg**3
        )

        def q_pow_r(q1, q2):
            return q1 ** q2

        self.assertRaises(ValueError, q_pow_r, 10.0 * pq.m, 10 * pq.J)
        self.assertRaises(ValueError, q_pow_r, 10.0 * pq.m, np.array([1, 2, 3]))

        self.assertQuantityEqual( (10 * pq.J) ** (2 * pq.J/pq.J) , 100 * pq.J**2 )

        # test rpow here
        self.assertRaises(ValueError, q_pow_r, 10.0, 10 * pq.J)

        self.assertQuantityEqual(10**(2*pq.J/pq.J), 100)

        def ipow(q1, q2):
            q1 -= q2
        self.assertRaises(ValueError, ipow, 1*pq.m, [1, 2])



"""
"""

from ._utils import _cd
from ..unitquantity import UnitConstant


Angstrom_star = UnitConstant(
    'Angstrom_star',
    _cd('Angstrom star'),
    symbol='A*',
    u_symbol='Å*'
)
d_220 = a_Si_220 = silicon_220_lattice_spacing = UnitConstant(
    'silicon_220_lattice_spacing',
    _cd('{220} lattice spacing of silicon'),
    symbol='d_220',
    u_symbol='d₂₂₀'
)
Cu_x_unit = UnitConstant(
    'Cu_x_unit',
    _cd('Cu x unit'),
    symbol='CuKalpha_1',
    u_symbol='CuKα₁'
)
a = a_Si_100 = lattice_parameter_of_silicon = UnitConstant(
    'lattice_parameter_of_silicon',
    _cd('lattice parameter of silicon')
)
Mo_x_unit = UnitConstant(
    'Mo_x_unit',
    _cd('Mo x unit'),
    symbol='MoKalpha_1',
    u_symbol='MoKα₁'
)

del UnitConstant, _cd



from .. import units as pq
from ..quantity import Quantity
from ..uncertainquantity import UncertainQuantity
from .common import TestCase
import numpy as np


class TestUncertainty(TestCase):

    def test_creation(self):
        a = UncertainQuantity(1, pq.m)
        self.assertQuantityEqual(a, 1*pq.m)
        self.assertQuantityEqual(a.uncertainty, 0*pq.m)
        a = UncertainQuantity([1, 1, 1], pq.m)
        self.assertQuantityEqual(a, [1,1,1]*pq.m)
        self.assertQuantityEqual(a.uncertainty, [0,0,0]*pq.m)
        a = UncertainQuantity([1, 1, 1], pq.m, [.1, .1, .1])
        self.assertQuantityEqual(a, [1, 1, 1] *pq.m)
        self.assertQuantityEqual(a.uncertainty, [0.1, 0.1, 0.1] *pq.m)
        self.assertRaises(ValueError, UncertainQuantity, [1,1,1], pq.m, 1)
        self.assertRaises(ValueError, UncertainQuantity, [1,1,1], pq.m, [1,1])

    def test_rescale(self):
        a = UncertainQuantity([1, 1, 1], pq.m, [.1, .1, .1])
        b = a.rescale(pq.ft)
        self.assertQuantityEqual(
            a.rescale('ft'),
            [3.2808399, 3.2808399, 3.2808399]*pq.ft
            )
        self.assertQuantityEqual(
            a.rescale('ft').uncertainty,
            [0.32808399, 0.32808399, 0.32808399]*pq.ft
            )

        seventy_km = Quantity(70, pq.km, dtype=np.float32)
        seven_km = Quantity(7, pq.km, dtype=np.float32)
        seventyish_km = UncertainQuantity(seventy_km, pq.km, seven_km, dtype=np.float32)
        self.assertTrue(seventyish_km.dtype == np.float32)
        in_meters = seventyish_km.rescale(pq.m)
        self.assertTrue(in_meters.dtype == seventyish_km.dtype)
        seventyish_km_rescaled_idempotent = seventyish_km.rescale(pq.km)
        self.assertTrue(seventyish_km_rescaled_idempotent.dtype == np.float32)
        self.assertQuantityEqual(seventyish_km + in_meters, 2*seventy_km)


    def test_set_uncertainty(self):
        a = UncertainQuantity([1, 2], 'm', [.1, .2])
        a.uncertainty = [1., 2.]*pq.m
        self.assertQuantityEqual(a.uncertainty, [1, 2]*pq.m)

        def set_u(q, u):
            q.uncertainty = u

        self.assertRaises(ValueError, set_u, a, 1)
        self.assertRaises(ValueError, set_u, a, [1,2])

    def test_uncertainquantity_multiply(self):
        a = UncertainQuantity([1, 2], 'm', [.1, .2])
        self.assertQuantityEqual(a*a, [1., 4.]*pq.m**2)
        self.assertQuantityEqual((a*a).uncertainty, [0.14142,0.56568]*pq.m**2)
        self.assertQuantityEqual(a*2, [2, 4]*pq.m)
        self.assertQuantityEqual((a*2).uncertainty, [0.2,0.4]*pq.m)

    def test_uncertainquantity_negative(self):
        a = UncertainQuantity([1, 2], 'm', [.1, .2])
        self.assertQuantityEqual(-a, [-1., -2.]*pq.m)
        self.assertQuantityEqual((-a).uncertainty, [0.1, 0.2]*pq.m)
        self.assertQuantityEqual(-a, a*-1)
        self.assertQuantityEqual((-a).uncertainty, (a*-1).uncertainty)

    def test_uncertainquantity_divide(self):
        a = UncertainQuantity([1, 2], 'm', [.1, .2])
        self.assertQuantityEqual(a/a, [1., 1.])
        self.assertQuantityEqual((a/a).uncertainty, [0.14142, 0.14142])
        self.assertQuantityEqual(a/pq.m, [1., 2.])
        self.assertQuantityEqual((a/pq.m).uncertainty, [0.1, 0.2])
        self.assertQuantityEqual(a/2, [0.5, 1.]*pq.m)
        self.assertQuantityEqual((a/2).uncertainty, [0.05, 0.1 ]*pq.m)
        self.assertQuantityEqual(1/a, [1., 0.5]/pq.m)
        self.assertQuantityEqual((1/a).uncertainty, [0.1, 0.05]/pq.m)

    def test_uncertaintity_mean(self):
        a = UncertainQuantity([1,2], 'm', [.1,.2])
        mean0 = np.sum(a)/np.size(a) # calculated traditionally
        mean1 = a.mean()        # calculated using this code
        self.assertQuantityEqual(mean0, mean1)

    def test_uncertaintity_nanmean(self):
        a = UncertainQuantity([1,2], 'm', [.1,.2])
        b = UncertainQuantity([1,2,np.nan], 'm', [.1,.2,np.nan])
        self.assertQuantityEqual(a.mean(),b.nanmean())

    def test_uncertainty_sqrt(self):
        a = UncertainQuantity([1,2], 'm', [.1,.2])
        self.assertQuantityEqual(a**0.5, a.sqrt())

    def test_uncertainty_nansum(self):
        uq = UncertainQuantity([1,2], 'm', [1,1])
        uq_nan = UncertainQuantity([1,2,np.nan], 'm', [1,1,np.nan])
        self.assertQuantityEqual(np.sum(uq), np.nansum(uq))
        self.assertQuantityEqual(np.sum(uq), uq_nan.nansum())

    def test_uncertainty_minmax_nan_arg(self):
        q = [[1, 2], [3, 4]] * pq.m        # quantity
        self.assertQuantityEqual(q.min(), 1*pq.m) # min
        self.assertQuantityEqual(q.max(), 4*pq.m) # max
        self.assertQuantityEqual(q.argmin(), 0) # argmin
        self.assertQuantityEqual(q.argmax(), 3) # argmax
        # uncertain quantity
        uq = UncertainQuantity([[1,2],[3,4]], pq.m, [[1,1],[1,1]])
        self.assertQuantityEqual(uq.min(), 1*pq.m) # min
        self.assertQuantityEqual(uq.max(), 4*pq.m) # max
        self.assertQuantityEqual(uq.argmin(), 0) # argmin
        self.assertQuantityEqual(uq.argmax(), 3) # argmax
        # now repeat the above with NaNs
        nanq = [[1, 2], [3, 4], [np.nan,np.nan]] * pq.m        # quantity
        nanuq = UncertainQuantity([[1,2],[3,4],[np.nan,np.nan]],
                               pq.m,
                               [[1,1],[1,1],[np.nan,np.nan]])
        self.assertQuantityEqual(nanq.nanmin(), 1*pq.m) # min
        self.assertQuantityEqual(nanq.nanmax(), 4*pq.m) # max
        self.assertQuantityEqual(nanq.nanargmin(), 0) # argmin
        self.assertQuantityEqual(nanq.nanargmax(), 3) # argmax
        self.assertQuantityEqual(nanuq.nanmin(), 1*pq.m) # min
        self.assertQuantityEqual(nanuq.nanmax(), 4*pq.m) # max
        self.assertQuantityEqual(nanuq.nanargmin(), 0) # argmin
        self.assertQuantityEqual(nanuq.nanargmax(), 3) # argmax
